from typing import Union, Optional
from logging import Logger
from pyspark.sql.types import StructType
from featurestorebundle.delta.TableExistenceChecker import TableExistenceChecker
from featurestorebundle.delta.EmptyDataFrameCreator import EmptyDataFrameCreator


class TableCreator:
    def __init__(
        self,
        logger: Logger,
        table_existence_checker: TableExistenceChecker,
        empty_dataframe_creator: EmptyDataFrameCreator,
    ):
        self.__logger = logger
        self.__table_existence_checker = table_existence_checker
        self.__empty_dataframe_creator = empty_dataframe_creator

    def create_if_not_exists(self, full_table_name: str, path: str, schema: StructType, partition_by: Optional[Union[str, list]] = None):
        partition_by = partition_by or []
        partition_by = [partition_by] if isinstance(partition_by, str) else partition_by

        if self.__table_existence_checker.exists(full_table_name):
            self.__logger.info(f"Table {full_table_name} already exists, creation skipped")
            return

        self.__logger.info(f"Creating new table {full_table_name} for {path}")

        df = self.__empty_dataframe_creator.create(schema)

        df.write.partitionBy(*partition_by).format("delta").option("path", path).saveAsTable(f"{full_table_name}")

        self.__logger.info(f"Table {full_table_name} successfully created")



from pyspark.sql import SparkSession
from featurestorebundle.db.TableNames import TableNames
from featurestorebundle.delta.TableCreator import TableCreator
from featurestorebundle.delta.metadata.schema import get_metadata_schema, get_metadata_pk_columns


class DeltaTableMetadataPreparer:
    def __init__(
        self,
        table_names: TableNames,
        table_creator: TableCreator,
        spark: SparkSession,
    ):
        self.__table_names = table_names
        self.__table_creator = table_creator
        self.__spark = spark

    def prepare(self, full_table_name: str, path: str):
        self.__spark.sql(f"CREATE DATABASE IF NOT EXISTS {self.__table_names.get_db_name()}")
        self.__table_creator.create_if_not_exists(full_table_name, path, get_metadata_schema(), get_metadata_pk_columns()[0].name)



from typing import Union
from logging import Logger
from pyspark.sql import SparkSession
from pyspark.sql import DataFrame
from featurestorebundle.db.TableNames import TableNames
from featurestorebundle.entity.Entity import Entity
from featurestorebundle.delta.TableExistenceChecker import TableExistenceChecker
from featurestorebundle.delta.EmptyDataFrameCreator import EmptyDataFrameCreator
from featurestorebundle.feature.reader.FeaturesReaderInterface import FeaturesReaderInterface
from featurestorebundle.delta.feature.schema import get_feature_store_initial_schema


class DeltaTableFeaturesReader(FeaturesReaderInterface):
    def __init__(
        self,
        logger: Logger,
        spark: SparkSession,
        table_names: TableNames,
        table_existence_checker: TableExistenceChecker,
        empty_dataframe_creator: EmptyDataFrameCreator,
    ):
        self.__logger = logger
        self.__spark = spark
        self.__table_names = table_names
        self.__table_existence_checker = table_existence_checker
        self.__empty_dataframe_creator = empty_dataframe_creator

    def read(self, entity: Union[Entity, str]) -> DataFrame:
        if isinstance(entity, Entity):
            entity = entity.name

        full_table_name = self.__table_names.get_features_full_table_name(entity)

        self.__logger.info(f"Reading features from table {full_table_name}")

        return self.__spark.read.table(full_table_name)

    def read_safe(self, entity: Entity) -> DataFrame:
        full_table_name = self.__table_names.get_features_full_table_name(entity.name)

        if not self.exists(entity):
            return self.__empty_dataframe_creator.create(get_feature_store_initial_schema(entity))

        return self.__spark.read.table(full_table_name)

    def exists(self, entity: Union[Entity, str]) -> bool:
        if isinstance(entity, Entity):
            entity = entity.name

        full_table_name = self.__table_names.get_features_full_table_name(entity)

        return self.__table_existence_checker.exists(full_table_name)



from featurestorebundle.delta.PathCreator import PathCreator
from featurestorebundle.delta.metadata.schema import get_metadata_schema, get_metadata_pk_columns


class DeltaPathMetadataPreparer:
    def __init__(
        self,
        path_creator: PathCreator,
    ):
        self.__path_creator = path_creator

    def prepare(self, path: str):
        self.__path_creator.create_if_not_exists(path, get_metadata_schema(), get_metadata_pk_columns()[0].name)



from pyspark.sql import SparkSession
from delta import DeltaTable


class PathExistenceChecker:
    def __init__(self, spark: SparkSession):
        self.__spark = spark

    def exists(self, path: str) -> bool:
        return DeltaTable.isDeltaTable(self.__spark, path)



from logging import Logger
from pyspark.sql import SparkSession
from delta import DeltaTable
from featurestorebundle.delta.feature.writer.DeltaFeaturesMergeConfig import DeltaFeaturesMergeConfig


class DeltaFeaturesDataHandler:
    def __init__(
        self,
        logger: Logger,
        spark: SparkSession,
    ):
        self.__logger = logger
        self.__spark = spark

    def merge_to_delta_table(self, full_table_name: str, config: DeltaFeaturesMergeConfig):
        delta_table = DeltaTable.forName(self.__spark, full_table_name)

        self.__logger.info(f"Writing features into {full_table_name}")

        self.__delta_merge(delta_table, config)

        self.__logger.info("Features write done")

    def merge_to_delta_path(self, path: str, config: DeltaFeaturesMergeConfig):
        delta_table = DeltaTable.forPath(self.__spark, path)

        self.__logger.info(f"Writing features into {path}")

        self.__delta_merge(delta_table, config)

        self.__logger.info("Features write done")

    def __delta_merge(self, delta_table: DeltaTable, config: DeltaFeaturesMergeConfig):  # noqa
        (
            delta_table.alias(config.target)
            .merge(config.data.alias(config.source), config.merge_condition)
            .whenMatchedUpdate(set=config.update_set)
            .whenNotMatchedInsert(values=config.insert_set)
            .execute()
        )



import pyspark.sql.types as t
from featurestorebundle.entity.Entity import Entity


def get_feature_store_initial_schema(entity: Entity):
    return t.StructType(
        [
            t.StructField(entity.id_column, entity.id_column_type, False),
            t.StructField(entity.time_column, entity.time_column_type, False),
            get_rainbow_table_hash_column(),
        ]
    )


def get_rainbow_table_hash_column():
    return t.StructField("features_hash", t.StringType(), False)


def get_rainbow_table_features_column():
    return t.StructField("computed_columns", t.ArrayType(t.StringType()), False)


def get_rainbow_table_schema():
    return t.StructType(
        [
            get_rainbow_table_hash_column(),
            get_rainbow_table_features_column(),
        ]
    )



from dataclasses import dataclass
from typing import List
from pyspark.sql.types import DataType


@dataclass(repr=True, frozen=True)
class Entity:
    name: str
    id_column: str
    id_column_type: DataType
    time_column: str
    time_column_type: DataType

    def get_primary_key(self) -> List[str]:
        return [self.id_column, self.time_column]



from pyspark.sql import SparkSession
from pyspark.sql import DataFrame
from delta import DeltaTable
from featurestorebundle.db.TableNames import TableNames
from featurestorebundle.delta.PathCreator import PathCreator
from featurestorebundle.delta.PathExistenceChecker import PathExistenceChecker
from featurestorebundle.delta.EmptyDataFrameCreator import EmptyDataFrameCreator
from featurestorebundle.delta.feature.schema import (
    get_rainbow_table_hash_column,
    get_rainbow_table_features_column,
    get_rainbow_table_schema,
)


class DeltaRainbowTableManager:
    def __init__(
        self,
        spark: SparkSession,
        table_names: TableNames,
        path_creator: PathCreator,
        path_existence_checker: PathExistenceChecker,
        empty_dataframe_creator: EmptyDataFrameCreator,
    ):
        self.__spark = spark
        self.__table_names = table_names
        self.__path_creator = path_creator
        self.__path_existence_checker = path_existence_checker
        self.__empty_dataframe_creator = empty_dataframe_creator

    def read(self, entity_name: str) -> DataFrame:
        return self.__spark.read.format("delta").load(self.__table_names.get_rainbow_path(entity_name))

    def read_safe(self, entity_name: str):
        if not self.exists(entity_name):
            return self.__empty_dataframe_creator.create(get_rainbow_table_schema())

        return self.__spark.read.format("delta").load(self.__table_names.get_rainbow_path(entity_name))

    def exists(self, entity_name: str) -> bool:
        return self.__path_existence_checker.exists(self.__table_names.get_rainbow_path(entity_name))

    def merge(self, entity_name: str, df: DataFrame):
        path = self.__table_names.get_rainbow_path(entity_name)

        self.__path_creator.create_if_not_exists(path, get_rainbow_table_schema())

        delta_table = DeltaTable.forPath(self.__spark, path)

        update_set = {get_rainbow_table_features_column().name: f"source.{get_rainbow_table_features_column().name}"}
        insert_set = {get_rainbow_table_hash_column().name: f"source.{get_rainbow_table_hash_column().name}", **update_set}
        merge_condition = f"target.{get_rainbow_table_hash_column().name} = source.{get_rainbow_table_hash_column().name}"

        (
            delta_table.alias("target")
            .merge(df.alias("source"), merge_condition)
            .whenMatchedUpdate(set=update_set)
            .whenNotMatchedInsert(values=insert_set)
            .execute()
        )



from typing import Optional
from daipecore.decorator.OutputDecorator import OutputDecorator
from injecta.container.ContainerInterface import ContainerInterface
from pyspark.sql import SparkSession
from pyspark.sql import DataFrame
from featurestorebundle.entity.Entity import Entity
from featurestorebundle.feature.FeatureTemplateMatcher import FeatureTemplateMatcher
from featurestorebundle.feature.FeatureTemplate import FeatureTemplate
from featurestorebundle.feature.FeatureList import FeatureList
from featurestorebundle.feature.FeaturesStorage import FeaturesStorage
from featurestorebundle.metadata.MetadataHTMLDisplayer import MetadataHTMLDisplayer


# pylint: disable=invalid-name
class feature(OutputDecorator):
    # pylint: disable=super-init-not-called
    def __init__(self, *args, entity: Entity, category: Optional[str] = None, features_storage: Optional[FeaturesStorage] = None):
        self._args = args
        self.__entity = entity
        self.__category = category
        self.__features_storage = features_storage

    def process_result(self, result: DataFrame, container: ContainerInterface):
        self.__check_primary_key_columns(result)

        feature_template_matcher: FeatureTemplateMatcher = container.get(FeatureTemplateMatcher)
        spark: SparkSession = container.get(SparkSession)
        feature_list = self.__prepare_features(feature_template_matcher, result, self._args)

        if not spark.sparkContext.getCheckpointDir():
            spark.sparkContext.setCheckpointDir("dbfs:/tmp/checkpoints")

        if container.get_parameters().featurestorebundle.metadata.display_in_notebook is True:
            metadata_html_displayer: MetadataHTMLDisplayer = container.get(MetadataHTMLDisplayer)
            metadata_html_displayer.display(feature_list.get_metadata_dicts())

        if container.get_parameters().featurestorebundle.orchestration.checkpoint_feature_results is True:
            result = result.checkpoint()

        if self.__features_storage is not None:
            self.__features_storage.add(result, feature_list)

    def __check_primary_key_columns(self, result: DataFrame):
        if self.__entity.id_column not in result.columns:
            raise Exception(f"{self.__entity.id_column} column is missing in the output dataframe")

        if self.__entity.time_column not in result.columns:
            raise Exception(f"{self.__entity.time_column} column is missing in the output dataframe")

    def __prepare_features(self, feature_template_matcher: FeatureTemplateMatcher, df: DataFrame, args: tuple) -> FeatureList:
        """
        @[foo]_feature(
            ("delayed_flights_pct_30d", "% of delayed flights in last 30 days"),
            ("early_flights_pct_30d", "% of flights landed ahead of time in last 30 days")
        )
        """
        feature_templates = [FeatureTemplate(*arg, category=self.__category) for arg in args]
        return feature_template_matcher.get_features(self.__entity, feature_templates, df)



import unittest
import warnings
from pyspark.sql import SparkSession
from pyspark.sql import DataFrame
from pyspark.sql import functions as f


class PySparkTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.spark = SparkSession.builder.master("local[1]").appName("PySparkTest").getOrCreate()
        cls.sc = cls.spark.sparkContext  # noqa

        warnings.filterwarnings("ignore", category=RuntimeWarning)
        warnings.filterwarnings("ignore", category=ResourceWarning)
        warnings.filterwarnings("ignore", category=DeprecationWarning)

    @classmethod
    def tearDownClass(cls):
        cls.sc.stop()  # noqa
        cls.spark.stop()  # noqa

    def compare_dataframes(self, df1: DataFrame, df2: DataFrame, sort_keys: list):
        df1_columns = sorted(df1.columns)
        df2_columns = sorted(df2.columns)

        df1 = df1.orderBy(sort_keys).select(df1_columns)
        df2 = df2.orderBy(sort_keys).select(df2_columns)

        self.assertEqual(df1.collect(), df2.collect())

    def delta_merge(self, target: DataFrame, source: DataFrame, primary_keys: list):  # noqa
        target_columns = target.columns
        source_columns = source.columns
        all_columns = list(set(target_columns) | set(source_columns))
        duplicate_columns = list((set(target_columns) & set(source_columns)) - set(primary_keys))
        unique_columns = list(set(all_columns) - set(primary_keys) - set(duplicate_columns))

        selection = []

        for col in all_columns:
            if col in unique_columns:
                selection.append(col)

            if col in duplicate_columns:
                selection.append(f.coalesce(f"source.{col}", f"target.{col}").alias(col))

        return target.alias("target").join(source.alias("source"), on=primary_keys, how="outer").select(*primary_keys, *selection)



from pyspark.sql import SparkSession
from featurestorebundle.db.TableNames import TableNames
from featurestorebundle.entity.Entity import Entity
from featurestorebundle.delta.TableCreator import TableCreator
from featurestorebundle.feature.FeatureList import FeatureList
from featurestorebundle.delta.feature.writer.DeltaFeaturesRegistrator import DeltaFeaturesRegistrator
from featurestorebundle.delta.feature.schema import get_feature_store_initial_schema


class DeltaTableFeaturesPreparer:
    def __init__(
        self,
        spark: SparkSession,
        table_names: TableNames,
        table_creator: TableCreator,
        features_registrator: DeltaFeaturesRegistrator,
    ):
        self.__spark = spark
        self.__table_names = table_names
        self.__table_creator = table_creator
        self.__features_registrator = features_registrator

    def prepare(self, full_table_name: str, path: str, entity: Entity, current_feature_list: FeatureList):
        self.__spark.sql(f"CREATE DATABASE IF NOT EXISTS {self.__table_names.get_db_name()}")
        self.__table_creator.create_if_not_exists(full_table_name, path, get_feature_store_initial_schema(entity), entity.time_column)
        self.__features_registrator.register(full_table_name, current_feature_list)



from featurestorebundle.feature.writer.FeaturesWriterInterface import FeaturesWriterInterface
from featurestorebundle.feature.FeaturesStorage import FeaturesStorage


class FeaturesWriter:
    def __init__(self, features_writer: FeaturesWriterInterface):
        self.__features_writer = features_writer

    def write(self, features_storage: FeaturesStorage):
        self.__features_writer.write(features_storage)



import unittest
import injecta.testing.services_tester as service_tester
from pyfonycore.bootstrap import bootstrapped_container


class FeatureStoreBundleTest(unittest.TestCase):
    def test_init(self):
        container = bootstrapped_container.init("test")

        service_tester.test_services(container)


if __name__ == "__main__":
    unittest.main()



from dataclasses import dataclass
from pyspark.sql import DataFrame
from typing import List


@dataclass(frozen=True)
class DatabricksFeatureStoreMergeConfig:
    data: DataFrame
    pk_columns: List[str]



from dataclasses import dataclass
from pyspark.sql import DataFrame
from typing import Dict


@dataclass(frozen=True)
class DeltaFeaturesMergeConfig:
    source: str
    target: str

    data: DataFrame

    merge_condition: str
    update_set: Dict[str, str]
    insert_set: Dict[str, str]



from pyspark.sql import SparkSession
from featurestorebundle.metadata.writer.MetadataWriterInterface import MetadataWriterInterface
from featurestorebundle.delta.metadata.writer.DeltaPathMetadataPreparer import DeltaPathMetadataPreparer
from featurestorebundle.delta.metadata.writer.DeltaMetadataHandler import DeltaMetadataHandler
from featurestorebundle.delta.metadata.schema import get_metadata_schema
from featurestorebundle.feature.FeatureList import FeatureList
from featurestorebundle.entity.Entity import Entity
from featurestorebundle.db.TableNames import TableNames


class DeltaPathMetadataWriter(MetadataWriterInterface):
    def __init__(
        self,
        spark: SparkSession,
        table_names: TableNames,
        path_preparer: DeltaPathMetadataPreparer,
        metadata_handler: DeltaMetadataHandler,
    ):
        self.__spark = spark
        self.__table_names = table_names
        self.__path_preparer = path_preparer
        self.__metadata_handler = metadata_handler

    def write(self, entity: Entity, feature_list: FeatureList):
        path = self.__table_names.get_metadata_path(entity.name)
        metadata_df = self.__spark.createDataFrame(feature_list.get_metadata(), get_metadata_schema())

        self.__path_preparer.prepare(path)
        self.__metadata_handler.merge_to_delta_path(path, metadata_df)



# pylint: disable=too-many-instance-attributes
class TableNames:
    def __init__(
        self,
        db_name: str,
        features_table_name_template: str,
        features_path_template: str,
        rainbow_path_template: str,
        metadata_table_name_template: str,
        metadata_path_template: str,
        entity_targets_table_name_template: str,
        entity_target_map_table_name: str,
        targets_table_name: str,
        entity_targets_path_template: str,
        entity_target_map_path: str,
        targets_path: str,
    ):
        self.__db_name = db_name
        self.__features_table_name_template = features_table_name_template
        self.__features_path_template = features_path_template
        self.__rainbow_path_template = rainbow_path_template
        self.__metadata_table_name_template = metadata_table_name_template
        self.__metadata_path_template = metadata_path_template
        self.__entity_targets_table_name_template = entity_targets_table_name_template
        self.__entity_target_map_table_name = entity_target_map_table_name
        self.__targets_table_name = targets_table_name
        self.__entity_targets_path_template = entity_targets_path_template
        self.__entity_target_map_path = entity_target_map_path
        self.__targets_path = targets_path

    def get_db_name(self) -> str:
        return self.__db_name

    def get_features_table_name(self, entity_name) -> str:
        return self.__replace_placeholders(self.__features_table_name_template, entity_name)

    def get_features_full_table_name(self, entity_name: str) -> str:
        return f"{self.get_db_name()}.{self.get_features_table_name(entity_name)}"

    def get_features_path(self, entity_name: str) -> str:
        return self.__replace_placeholders(self.__features_path_template, entity_name)

    def get_metadata_table_name(self, entity_name: str) -> str:
        return self.__replace_placeholders(self.__metadata_table_name_template, entity_name)

    def get_metadata_full_table_name(self, entity_name: str) -> str:
        return f"{self.get_db_name()}.{self.get_metadata_table_name(entity_name)}"

    def get_metadata_path(self, entity_name: str) -> str:
        return self.__replace_placeholders(self.__metadata_path_template, entity_name)

    def get_rainbow_path(self, entity_name: str) -> str:
        return self.__replace_placeholders(self.__rainbow_path_template, entity_name)

    def get_entity_targets_table_name(self, entity_name: str) -> str:
        return self.__replace_placeholders(self.__entity_targets_table_name_template, entity_name)

    def get_entity_target_map_table_name(self) -> str:
        return self.__entity_target_map_table_name

    def get_targets_table_name(self) -> str:
        return self.__targets_table_name

    def get_entity_targets_full_table_name(self, entity_name: str) -> str:
        return f"{self.get_db_name()}.{self.get_entity_targets_table_name(entity_name)}"

    def get_entity_target_map_full_table_name(self) -> str:
        return f"{self.get_db_name()}.{self.get_entity_target_map_table_name()}"

    def get_targets_full_table_name(self) -> str:
        return f"{self.get_db_name()}.{self.get_targets_table_name()}"

    def get_entity_targets_path(self, entity_name: str) -> str:
        return self.__replace_placeholders(self.__entity_targets_path_template, entity_name)

    def get_entity_target_map_path(self) -> str:
        return self.__entity_target_map_path

    def get_targets_path(self) -> str:
        return self.__targets_path

    def __replace_placeholders(self, template: str, entity_name: str) -> str:
        return template.format(entity=entity_name)



import pyspark.sql.types as t
from featurestorebundle.entity.Entity import Entity


def get_target_id_column_name():
    return "target_id"


def get_id_column_name():
    return "id_column"


def get_time_column_name():
    return "time_column"


def get_entity_targets_schema(entity: Entity):
    return [
        t.StructField("id_column", entity.id_column_type, False),
        t.StructField("time_column", entity.time_column_type, False),
        t.StructField("target_id", t.StringType(), False),
    ]


def get_entity_target_map_schema():
    return [
        t.StructField("entity", t.StringType(), False),
        t.StructField("target_id", t.StringType(), False),
    ]


def get_targets_schema():
    return [
        t.StructField("target_id", t.StringType(), False),
        t.StructField("description", t.StringType(), True),
    ]



from typing import List, Dict, Union

from featurestorebundle.feature.Feature import Feature


class FeatureList:
    def __init__(self, features: List[Feature]):
        self.__features = features

    def get_all(self) -> List[Feature]:
        return self.__features

    def empty(self):
        return not self.__features

    def get_names(self):
        return [feature.name for feature in self.__features]

    def get_unregistered(self, registered_feature_names: List[str]) -> "FeatureList":
        def registered(instance: Feature, registered_names: List[str]):
            return instance.name in registered_names

        return FeatureList([feature for feature in self.get_all() if not registered(feature, registered_feature_names)])

    def merge(self, new_feature_list: "FeatureList") -> "FeatureList":
        return FeatureList(self.__features + new_feature_list.get_all())

    def get_metadata(self) -> List[List[Union[Dict[str, str], str]]]:
        return [feature.get_metadata_list() for feature in self.__features]

    def get_metadata_dicts(self) -> List[Dict[str, Union[str, Dict[str, str]]]]:
        return [feature.get_metadata_dict() for feature in self.__features]



from abc import ABC, abstractmethod
from typing import Union
from pyspark.sql import DataFrame
from featurestorebundle.entity.Entity import Entity


class FeaturesReaderInterface(ABC):
    @abstractmethod
    def read(self, entity: Union[Entity, str]) -> DataFrame:
        pass

    @abstractmethod
    def read_safe(self, entity: Entity) -> DataFrame:
        pass

    @abstractmethod
    def exists(self, entity: Union[Entity, str]) -> bool:
        pass



from pyspark.sql import SparkSession
from pyspark.sql import DataFrame
from logging import Logger
from featurestorebundle.db.TableNames import TableNames
from featurestorebundle.target.reader.TargetsReaderInterface import TargetsReaderInterface


class DeltaPathTargetsReader(TargetsReaderInterface):
    def __init__(
        self,
        logger: Logger,
        table_names: TableNames,
        spark: SparkSession,
    ):
        self.__logger = logger
        self.__spark = spark
        self.__table_names = table_names

    def read(self, entity: str) -> DataFrame:
        path = self.__table_names.get_entity_targets_path(entity)

        self.__logger.info(f"Reading targets for entity {entity} from path {path}")

        return self.__spark.read.format("delta").load(path)



from pyspark.sql import Column, functions as f
from typing import Any, List, Callable

from pyspark.sql import DataFrame


__HOUR = 60 * 60
__DAY = 24 * __HOUR
__WEEK = 7 * __DAY

__PERIODS = {
    "h": __HOUR,
    "d": __DAY,
    "w": __WEEK,
}

# pylint: disable=invalid-name
__time_window_column_template = "is_time_window_{time_window}"


def _is_past_time_window(target_date: Column, window_argument_date: Column, time_window) -> Column:
    period = __PERIODS[time_window[-1]] * int(time_window[:-1])
    delta = target_date - window_argument_date
    return (0 <= delta) & (delta <= period)


def __resolve_column_type(df: DataFrame, window_col: str) -> Column:
    dtypes = dict(df.dtypes)

    if window_col not in dtypes:
        raise ValueError(f"Column {window_col} not found in dataframe.")

    dtype = dtypes[window_col]

    if dtype == "date":
        return f.to_timestamp(f.col(window_col)).cast("long")

    if dtype == "timestamp":
        return f.col(window_col).cast("long")

    raise TypeError(f"Column {window_col} is of unsupported type '{dtype}'. Must be either 'date' or 'timestamp'.")


def __with_time_windows(
    df: DataFrame,
    window_col: str,
    target_date_column: Column,
    time_windows: List,
    is_time_window_function: Callable,
    time_window_column_template: str,
) -> DataFrame:
    window_argument_date = __resolve_column_type(df, window_col)
    target_date = target_date_column.cast("long")

    time_window_columns = [
        is_time_window_function(target_date, window_argument_date, time_window).alias(
            time_window_column_template.format(time_window=time_window)
        )
        for time_window in time_windows
    ]

    return df.select("*", *time_window_columns)


def __windowed(time_window_column_template: str, column: Column, time_window: str, default_value: Any = 0) -> Column:
    time_window_col_name = time_window_column_template.format(time_window=time_window)
    return f.when(f.col(time_window_col_name), column).otherwise(f.lit(default_value))


def with_time_windows(df: DataFrame, window_col: str, target_date_column: Column, time_windows: List) -> DataFrame:
    return __with_time_windows(df, window_col, target_date_column, time_windows, _is_past_time_window, __time_window_column_template)


def windowed(column: Column, time_window: str, default_value: Any = 0) -> Column:
    return __windowed(__time_window_column_template, column, time_window, default_value)



from featurestorebundle.entity.Entity import Entity
from featurestorebundle.delta.PathCreator import PathCreator
from featurestorebundle.feature.FeatureList import FeatureList
from featurestorebundle.delta.feature.writer.DeltaFeaturesRegistrator import DeltaFeaturesRegistrator
from featurestorebundle.delta.feature.schema import get_feature_store_initial_schema


class DeltaPathFeaturesPreparer:
    def __init__(
        self,
        path_creator: PathCreator,
        features_registrator: DeltaFeaturesRegistrator,
    ):
        self.__path_creator = path_creator
        self.__features_registrator = features_registrator

    def prepare(self, path: str, entity: Entity, current_feature_list: FeatureList):
        self.__path_creator.create_if_not_exists(path, get_feature_store_initial_schema(entity), entity.time_column)
        self.__features_registrator.register(f"delta.`{path}`", current_feature_list)



import hashlib
from functools import reduce
from typing import List, Tuple
from logging import Logger
from pyspark.sql import DataFrame
from pyspark.sql import functions as f
from pyspark.sql.window import Window
from featurestorebundle.feature.FeaturesStorage import FeaturesStorage
from featurestorebundle.feature.FeatureList import FeatureList
from featurestorebundle.delta.feature.schema import get_rainbow_table_hash_column, get_rainbow_table_features_column


class FeaturesPreparer:
    def __init__(
        self,
        logger: Logger,
        join_method: str,
        join_batch_size: int,
        checkpoint_after_join: bool,
        checkpoint_before_merge: bool,
    ):
        self.__logger = logger
        self.__join_method = join_method
        self.__join_batch_size = join_batch_size
        self.__checkpoint_after_join = checkpoint_after_join
        self.__checkpoint_before_merge = checkpoint_before_merge

    def prepare(self, features_storage: FeaturesStorage, feature_store: DataFrame, rainbow_table: DataFrame) -> Tuple[DataFrame, DataFrame]:
        entity = features_storage.entity
        feature_list = features_storage.feature_list

        base_dataframe = self.__prepare_base_dataframe(features_storage, feature_store, rainbow_table)

        if self.__checkpoint_before_merge:
            self.__logger.info("Checkpointing features data before merge")

            base_dataframe = base_dataframe.checkpoint()

            self.__logger.info("Checkpointing done")

        features_data = base_dataframe.select(
            entity.id_column,
            entity.time_column,
            get_rainbow_table_hash_column().name,
            *feature_list.get_names(),
        )

        rainbow_data = base_dataframe.select(
            get_rainbow_table_hash_column().name,
            get_rainbow_table_features_column().name,
        ).distinct()

        return features_data, rainbow_data

    def __prepare_base_dataframe(self, features_storage: FeaturesStorage, feature_store: DataFrame, rainbow_table: DataFrame) -> DataFrame:
        if not features_storage.results:
            raise Exception("There are no features to write.")

        entity = features_storage.entity
        feature_list = features_storage.feature_list
        results = features_storage.results
        pk_columns = [entity.id_column, entity.time_column]
        technical_columns = pk_columns + [get_rainbow_table_hash_column().name]

        registered_features = {col for col in feature_store.columns if col not in technical_columns}
        incoming_features = {*feature_list.get_names()}

        joined_results = self.__join_results(results, pk_columns)

        if registered_features == incoming_features:
            self.__logger.debug("Optimization: schema did not change")

            return self.__compute_new_hashes_optimized(joined_results, feature_list)

        joined_results = joined_results.withColumn("new_columns", f.array(*map(f.lit, feature_list.get_names())))

        return self.__compute_new_hashes(joined_results, feature_store, rainbow_table)

    def __compute_new_hashes(self, joined_results: DataFrame, feature_store: DataFrame, rainbow_table: DataFrame) -> DataFrame:  # noqa
        pk_columns = feature_store.columns[0:2]
        technical_columns = feature_store.columns[0:3]

        return (
            joined_results.join(feature_store.select(technical_columns), on=pk_columns, how="left")
            .join(rainbow_table, on=get_rainbow_table_hash_column().name, how="left")
            .withColumn("computed_columns", f.when(f.col("computed_columns").isNull(), f.array()).otherwise(f.col("computed_columns")))
            .withColumn("columns_union", f.array_sort(f.array_union("computed_columns", "new_columns")))
            .withColumn("new_features_hash", f.md5(f.concat_ws("`", "columns_union")))
            .drop("features_hash", "computed_columns")
            .withColumnRenamed("new_features_hash", "features_hash")
            .withColumnRenamed("columns_union", "computed_columns")
        )

    def __compute_new_hashes_optimized(self, joined_results: DataFrame, feature_list: FeatureList) -> DataFrame:  # noqa
        id_column = joined_results.columns[0]
        time_column = joined_results.columns[1]

        features_hash = hashlib.md5("`".join(sorted(feature_list.get_names())).encode()).hexdigest()

        return joined_results.select(
            id_column,
            time_column,
            f.lit(features_hash).alias(get_rainbow_table_hash_column().name),
            f.lit(f.array_sort(f.array(*map(f.lit, feature_list.get_names())))).alias(get_rainbow_table_features_column().name),
            *feature_list.get_names(),
        )

    def __join_results(self, results: List[DataFrame], pk_columns: List[str]):
        if self.__join_method == "left_with_checkpointing":
            join_method = self.__join_dataframes_using_left_join

        elif self.__join_method == "union_with_window":
            join_method = self.__join_dataframes_using_union_and_window

        else:
            raise Exception("Invalid join method")

        joined_results = join_method(results, pk_columns)

        if self.__checkpoint_after_join:
            self.__logger.info("Checkpointing features data after join")

            joined_results = joined_results.checkpoint()

            self.__logger.info("Checkpointing done")

        return joined_results

    def __join_dataframes_using_left_join(self, dfs: List[DataFrame], join_columns: List[str]) -> DataFrame:
        join_batch_counter = 0
        id_dataframes = [df.select(join_columns) for df in dfs]
        unique_ids_df = reduce(lambda df1, df2: df1.unionByName(df2), id_dataframes).distinct().cache()
        joined_df = unique_ids_df

        for df in dfs:
            join_batch_counter += 1
            joined_df = joined_df.join(df, on=join_columns, how="left")

            if join_batch_counter == self.__join_batch_size:
                joined_df = joined_df.checkpoint()
                join_batch_counter = 0

        return joined_df

    def __join_dataframes_using_union_and_window(self, dfs: List[DataFrame], join_columns: List[str]) -> DataFrame:  # noqa
        window = Window.partitionBy(*join_columns).rowsBetween(Window.unboundedPreceding, Window.unboundedFollowing)
        union_df = reduce(lambda df1, df2: df1.unionByName(df2, allowMissingColumns=True), dfs)
        features = [col for col in union_df.columns if col not in join_columns]

        return (
            union_df.select(
                *join_columns,
                *[f.first(feature, ignorenulls=True).over(window).alias(feature) for feature in features],
            )
            .groupBy(join_columns)
            .agg(*[f.first(feature).alias(feature) for feature in features])
        )



import unittest
from featurestorebundle.feature.FeatureTemplate import FeatureTemplate
from featurestorebundle.feature.FeaturePattern import FeaturePattern


class FeaturePatternTest(unittest.TestCase):
    def test_simple(self):
        template = FeatureTemplate("test_feature", "test description")
        pattern = FeaturePattern(template)
        match = pattern.get_match("test_feature")

        self.assertIsNotNone(match)

    def test_placeholders(self):
        template = FeatureTemplate("test_{arg1}_{arg2}_something", "test description")
        pattern = FeaturePattern(template)
        match = pattern.get_match("test_hello_world_something")

        self.assertIsNotNone(match)
        self.assertDictEqual({"arg1": "hello", "arg2": "world"}, pattern.get_groups_as_dict(match))

    def test_appedix(self):
        template = FeatureTemplate("test_something", "test description")
        pattern = FeaturePattern(template)
        match = pattern.get_match("test_something_else")

        self.assertIsNone(match)

        template = FeatureTemplate("test_{arg1}_{arg2}_something", "test description")
        pattern = FeaturePattern(template)
        match = pattern.get_match("test_hello_world_something_else")

        self.assertIsNone(match)


if __name__ == "__main__":
    unittest.main()



from typing import List, Optional
from datetime import date
from pyspark.sql import DataFrame
from featurestorebundle.feature.reader.FeaturesReaderInterface import FeaturesReaderInterface
from featurestorebundle.metadata.reader.MetadataReaderInterface import MetadataReaderInterface
from featurestorebundle.target.reader.TargetsReaderInterface import TargetsReaderInterface
from featurestorebundle.delta.feature.DeltaRainbowTableManager import DeltaRainbowTableManager
from featurestorebundle.delta.feature.FeaturesFilteringManager import FeaturesFilteringManager
from featurestorebundle.delta.target.TargetsFilteringManager import TargetsFilteringManager


class FeatureStore:
    def __init__(
        self,
        features_reader: FeaturesReaderInterface,
        metadata_reader: MetadataReaderInterface,
        targets_reader: TargetsReaderInterface,
        rainbow_table_manager: DeltaRainbowTableManager,
        features_filtering_manager: FeaturesFilteringManager,
        targets_filtering_manager: TargetsFilteringManager,
    ):
        self.__features_reader = features_reader
        self.__metadata_reader = metadata_reader
        self.__targets_reader = targets_reader
        self.__rainbow_table_manager = rainbow_table_manager
        self.__features_filtering_manager = features_filtering_manager
        self.__targets_filtering_manager = targets_filtering_manager

    def get_latest(self, entity: str, features: Optional[List[str]] = None) -> DataFrame:
        features = features or []
        feature_store = self.__features_reader.read(entity)
        rainbow_table = self.__rainbow_table_manager.read(entity)

        return self.__features_filtering_manager.get_latest(feature_store, rainbow_table, features)

    def get_for_target(
        self,
        entity: str,
        target: str,
        target_date_from: Optional[date] = None,
        target_date_to: Optional[date] = None,
        time_diff: Optional[str] = None,
        features: Optional[List[str]] = None,
        skip_incomplete_rows: bool = False,
    ) -> DataFrame:
        features = features or []
        feature_store = self.__features_reader.read(entity)
        target_store = self.__targets_reader.read(entity)
        rainbow_table = self.__rainbow_table_manager.read(entity)
        targets = self.__targets_filtering_manager.get_targets(target_store, target, target_date_from, target_date_to, time_diff)

        return self.__features_filtering_manager.get_for_target(feature_store, rainbow_table, targets, features, skip_incomplete_rows)

    def get_targets(
        self,
        entity: str,
        target: str,
        date_from: Optional[date] = None,
        date_to: Optional[date] = None,
        time_diff: Optional[str] = None,
    ) -> DataFrame:
        target_store = self.__targets_reader.read(entity)

        return self.__targets_filtering_manager.get_targets(target_store, target, date_from, date_to, time_diff)

    def get_metadata(
        self,
        entity: Optional[str] = None,
    ) -> DataFrame:
        return self.__metadata_reader.read(entity)



from typing import Optional

from box import Box
from daipecore.widgets.Widgets import Widgets


class WidgetsFactory:
    def __init__(self, defaults: Box, entities: Box, widgets: Widgets):
        self.__defaults = defaults
        self.__entities = entities
        self.__widgets = widgets

    def create(self, default_time_window: Optional[str] = None):
        default_time_window = self.__defaults.time_windows if default_time_window is None else default_time_window

        self.__widgets.remove_all()

        self.create_for_entity()

        self.create_target_name()

        if self.__widgets.get_value("target_name") == "<no target>":
            self.create_for_run_date(default_time_window)
        else:
            self.create_for_target(default_time_window)

    def create_for_entity(self):
        entities_list = list(self.__entities)

        if len(entities_list) > 1:
            self.__widgets.add_select("entity", entities_list, default_value=entities_list[0])

    def create_for_run_date(self, default_time_window: Optional[str] = None):
        default_time_window = self.__defaults.time_windows if default_time_window is None else default_time_window

        self.__widgets.add_text("run_date", self.__defaults.run_date)

        self.create_time_window(default_time_window)

    def create_for_target(self, default_time_window: Optional[str] = None):
        default_time_window = self.__defaults.time_windows if default_time_window is None else default_time_window

        self.create_time_window(default_time_window)

        self.__widgets.add_text("target_date_from", self.__defaults.target_date_from)

        self.__widgets.add_text("target_date_to", self.__defaults.target_date_to)

        self.__widgets.add_text("number_of_time_units", self.__defaults.number_of_time_units)

    def create_target_name(self):
        self.__widgets.add_select(
            "target_name",
            [
                "<no target>",
                "Date of report of change of female name",
                "First mortgage payment",
                "First loan payment",
                "Date of opening the investment account",
                "Last payment date for the veterinarian",
                "First payment date of unemployment benefits",
                "First payment date of maternity leave",
                "First payment date of alimony",
                "First payment date of paternity leave",
            ],
            "<no target>",
        )

    def create_time_window(self, default_time_window: str):
        self.__widgets.add_text("time_window", default_time_window)



from daipecore.decorator.ContainerManager import ContainerManager
from featurestorebundle.entity.EntityGetter import EntityGetter
from featurestorebundle.entity.Entity import Entity


def get_entity() -> Entity:
    container = ContainerManager.get_container()
    entity_getter: EntityGetter = container.get(EntityGetter)

    return entity_getter.get()



from typing import Union
from logging import Logger
from pyspark.sql import DataFrame
from featurestorebundle.db.TableNames import TableNames
from featurestorebundle.entity.Entity import Entity
from featurestorebundle.delta.TableExistenceChecker import TableExistenceChecker
from featurestorebundle.delta.EmptyDataFrameCreator import EmptyDataFrameCreator
from featurestorebundle.feature.reader.FeaturesReaderInterface import FeaturesReaderInterface
from featurestorebundle.delta.feature.schema import get_feature_store_initial_schema


class DatabricksFeatureStoreReader(FeaturesReaderInterface):
    def __init__(
        self,
        logger: Logger,
        table_names: TableNames,
        table_existence_checker: TableExistenceChecker,
        empty_dataframe_creator: EmptyDataFrameCreator,
    ):
        self.__logger = logger
        self.__table_names = table_names
        self.__table_existence_checker = table_existence_checker
        self.__empty_dataframe_creator = empty_dataframe_creator

    def read(self, entity: Union[Entity, str]) -> DataFrame:
        if isinstance(entity, Entity):
            entity = entity.name

        full_table_name = self.__table_names.get_features_full_table_name(entity)

        self.__logger.info(f"Reading features from Databricks Feature Store table {full_table_name}")

        return self.__read_from_databricks_feature_store(full_table_name)

    def read_safe(self, entity: Entity) -> DataFrame:
        full_table_name = self.__table_names.get_features_full_table_name(entity.name)

        if not self.exists(entity):
            return self.__empty_dataframe_creator.create(get_feature_store_initial_schema(entity))

        return self.__read_from_databricks_feature_store(full_table_name)

    def exists(self, entity: Union[Entity, str]) -> bool:
        if isinstance(entity, Entity):
            entity = entity.name

        full_table_name = self.__table_names.get_features_full_table_name(entity)

        return self.__table_existence_checker.exists(full_table_name)

    def __read_from_databricks_feature_store(self, full_table_name: str) -> DataFrame:
        try:
            from databricks import feature_store  # pyre-ignore[21] pylint: disable=import-outside-toplevel

        except ImportError as exception:
            raise Exception("Cannot import Databricks Feature Store, you need to use ML cluster with DBR 8.3+") from exception

        dbx_feature_store_client = feature_store.FeatureStoreClient()

        return dbx_feature_store_client.read_table(full_table_name)



import re
from typing import Optional
from datetime import date
from pyspark.sql import DataFrame
from pyspark.sql import Column
from pyspark.sql import functions as f
from featurestorebundle.delta.target.schema import get_target_id_column_name, get_id_column_name, get_time_column_name


class TargetsFilteringManager:
    def get_targets(
        self,
        targets: DataFrame,
        target_id: str,
        date_from: Optional[date] = None,
        date_to: Optional[date] = None,
        time_diff: Optional[str] = None,
    ) -> DataFrame:
        df = targets.filter(f.col(get_target_id_column_name()) == target_id)

        if date_from:
            df = df.filter(f.col(get_time_column_name()) >= date_from)

        if date_to:
            df = df.filter(f.col(get_time_column_name()) <= date_to)

        if time_diff:
            df = df.withColumn(get_time_column_name(), self.__shift_time_column(time_diff))

        return df.select(get_id_column_name(), get_time_column_name())

    def __shift_time_column(self, time_diff: str) -> Column:
        matches = re.match(r"([+-]?[0-9]+)([smhdw])", time_diff)

        if not matches:
            raise Exception("Invalid time format try something like '7d' for seven days")

        periods = {
            "s": "SECONDS",
            "m": "MINUTES",
            "h": "HOURS",
            "d": "DAYS",
            "w": "WEEKS",
        }

        integer_part = matches[1]
        period_part = matches[2]

        return f.col(get_time_column_name()) + f.expr(f"INTERVAL {integer_part} {periods[period_part]}")



from pyspark.sql import DataFrame
from featurestorebundle.entity.Entity import Entity
from featurestorebundle.feature.FeatureList import FeatureList
from featurestorebundle.delta.feature.schema import get_rainbow_table_hash_column


class FeaturesValidator:
    def validate(self, entity: Entity, features_data: DataFrame, feature_list: FeatureList):
        technical_cols = [entity.id_column, entity.time_column, get_rainbow_table_hash_column().name]

        data_column_names = list(
            map(
                lambda field: field.name,
                filter(lambda field: field.name not in technical_cols, features_data.schema.fields),
            )
        )

        feature_names = feature_list.get_names()

        if data_column_names != feature_names:
            raise Exception(f"Dataframe columns of size ({len(data_column_names)}) != features matched of size ({len(feature_names)})")



from pyspark.sql import DataFrame
from featurestorebundle.feature.FeaturesStorage import FeaturesStorage


class FeaturesPreparer:
    def prepare(self, features_storage: FeaturesStorage) -> DataFrame:
        join_batch_size = 10
        batch_counter = 0

        if not features_storage.results:
            raise Exception("There are no features to write.")

        pk_columns = [features_storage.entity.id_column, features_storage.entity.time_column]

        id_dataframes = [df.select(pk_columns) for df in features_storage.results]

        unique_ids_df = id_dataframes[0]

        for df in id_dataframes[1:]:
            unique_ids_df = unique_ids_df.unionByName(df)

        unique_ids_df = unique_ids_df.distinct()
        joined_df = unique_ids_df.cache()

        for df in features_storage.results:
            batch_counter += 1

            joined_df = joined_df.join(df, on=pk_columns, how="left")

            if batch_counter == join_batch_size:
                joined_df = joined_df.persist()
                batch_counter = 0

        return joined_df



import pyspark.sql.types as t


def get_metadata_pk_columns():
    return [
        t.StructField("entity", t.StringType(), False),
        t.StructField("feature", t.StringType(), False),
    ]


def get_metadata_columns():
    return [
        t.StructField("description", t.StringType(), True),
        t.StructField("extra", t.MapType(t.StringType(), t.StringType(), True)),
        t.StructField("feature_template", t.StringType(), True),
        t.StructField("category", t.StringType(), True),
        t.StructField("dtype", t.StringType(), True),
    ]


def get_metadata_schema():
    return t.StructType(get_metadata_pk_columns() + get_metadata_columns())


def get_metadata_string_schema():
    return ", ".join([f"{field.name} {field.dataType.simpleString()}" for field in get_metadata_schema()])



from pyspark.sql import SparkSession


class DummySparkFactory:
    def create(self):
        class DummySparkSession(SparkSession):
            # pylint: disable=super-init-not-called
            def __init__(self, *args):
                pass

            @property
            def _sc(self):
                return DummySparkContext()

        class DummySparkContext:
            def getCheckpointDir(self):  # noqa pylint: disable=invalid-name
                return ""

            def setCheckpointDir(self, dirName):  # noqa pylint: disable=invalid-name
                pass

        return DummySparkSession()



from abc import ABC, abstractmethod
from typing import Optional
from pyspark.sql import DataFrame


class MetadataReaderInterface(ABC):
    @abstractmethod
    def read(self, entity_name: Optional[str] = None) -> DataFrame:
        pass



from logging import Logger
from pyspark.sql import SparkSession
from pyspark.sql import DataFrame
from delta import DeltaTable
from featurestorebundle.delta.metadata.schema import get_metadata_pk_columns, get_metadata_columns


class DeltaMetadataHandler:
    def __init__(
        self,
        logger: Logger,
        spark: SparkSession,
    ):
        self.__logger = logger
        self.__spark = spark

    def merge_to_delta_table(self, full_table_name: str, metadata_df: DataFrame):
        delta_table = DeltaTable.forName(self.__spark, full_table_name)

        self.__logger.info(f"Writing metadata into {full_table_name}")

        self.__delta_merge(delta_table, metadata_df)

        self.__logger.info("Metadata write done")

    def merge_to_delta_path(self, path: str, metadata_df: DataFrame):
        delta_table = DeltaTable.forPath(self.__spark, path)

        self.__logger.info(f"Writing metadata into {path}")

        self.__delta_merge(delta_table, metadata_df)

        self.__logger.info("Metadata write done")

    def __delta_merge(self, delta_table: DeltaTable, metadata_df: DataFrame):  # noqa
        update_set = {col.name: f"source.{col.name}" for col in get_metadata_columns()}
        insert_set = {**{col.name: f"source.{col.name}" for col in get_metadata_pk_columns()}, **update_set}
        merge_condition = " AND ".join(f"target.{col.name} = source.{col.name}" for col in get_metadata_pk_columns())

        (
            delta_table.alias("target")
            .merge(metadata_df.alias("source"), merge_condition)
            .whenMatchedUpdate(set=update_set)
            .whenNotMatchedInsert(values=insert_set)
            .execute()
        )



from typing import Optional


class FeatureTemplate:
    def __init__(self, name_template: str, description_template: str, category: Optional[str] = None):

        self.__name_template = name_template
        self.__description_template = description_template
        self.__category = category

    @property
    def name_template(self):
        return self.__name_template

    @property
    def description_template(self):
        return self.__description_template

    @property
    def category(self):
        return self.__category



from featurestorebundle.delta.feature.writer.DeltaFeaturesMergeConfig import DeltaFeaturesMergeConfig
from featurestorebundle.entity.Entity import Entity
from pyspark.sql import DataFrame
from typing import List, Dict


class DeltaFeaturesMergeConfigGenerator:
    def generate(
        self,
        entity: Entity,
        features_data: DataFrame,
        pk_columns: List[str],
    ) -> DeltaFeaturesMergeConfig:
        technical_cols = [entity.id_column, entity.time_column]
        data_column_names = list(
            map(
                lambda field: field.name,
                filter(lambda field: field.name not in technical_cols, features_data.schema.fields),
            )
        )

        update_set = self.__build_set(data_column_names)

        if entity.time_column not in pk_columns:
            update_set[entity.time_column] = self.__wrap_source(entity.time_column)

        insert_set = {**update_set, **self.__build_set(technical_cols)}
        merge_condition = " AND ".join(f"{self.__wrap_target(pk)} = {self.__wrap_source(pk)}" for pk in pk_columns)

        return DeltaFeaturesMergeConfig(
            source="source",
            target="target",
            data=features_data,
            merge_condition=merge_condition,
            update_set=update_set,
            insert_set=insert_set,
        )

    def __wrap_source(self, column: str) -> str:
        return self.__wrap("source", column)

    def __wrap_target(self, column: str) -> str:
        return self.__wrap("target", column)

    def __build_set(self, columns: List[str]) -> Dict[str, str]:
        return {column: self.__wrap_source(column) for column in columns}

    def __wrap(self, alias: str, column: str) -> str:  # noqa
        return f"{alias}.{column}"



import os
from pyspark.sql import types as t
from daipecore.decorator.DecoratedDecorator import DecoratedDecorator
from daipecore.decorator.notebook_function import notebook_function
from featurestorebundle.entity.Entity import Entity
from featurestorebundle.feature.FeaturesStorage import FeaturesStorage
from featurestorebundle.notebook.decorator.feature import feature

os.environ["APP_ENV"] = "test"


class FakeSchema:
    json = {
        "fields": [
            {"name": "client_id", "type": "long"},
            {"name": "run_date", "type": "long"},
            {"name": "my_sample_feature", "type": "long"},
        ],
    }

    # pylint: disable=invalid-name
    @staticmethod
    def jsonValue():
        return FakeSchema.json


class FakeResult:

    columns = ["client_id", "run_date", "my_sample_feature"]

    def __init__(self, value):
        self.value = value
        self.schema = FakeSchema


entity = Entity(
    name="client_test",
    id_column="client_id",
    id_column_type=t.StringType(),
    time_column="run_date",
    time_column_type=t.DateType(),
)

features_storage = FeaturesStorage(entity)


# pylint: disable=invalid-name
@DecoratedDecorator
class client_feature(feature):
    def __init__(self, *args, category=None):
        super().__init__(*args, entity=entity, category=category, features_storage=features_storage)


expected_value = FakeResult("not_a_real_dataframe")

try:

    @notebook_function()
    @client_feature(("my_sample_feature", "my_sample_description"), category="test")
    def my_sample_feature():
        return expected_value


except ModuleNotFoundError as e:
    assert str(e) == "No module named 'IPython'"

assert expected_value == features_storage.results[0]



from abc import ABC, abstractmethod
from featurestorebundle.entity.Entity import Entity
from featurestorebundle.feature.FeatureList import FeatureList


class MetadataWriterInterface(ABC):
    @abstractmethod
    def write(self, entity: Entity, feature_list: FeatureList):
        pass



from typing import List
from pyspark.sql import functions as f
from pyspark.sql import DataFrame
from featurestorebundle.delta.feature.schema import get_rainbow_table_hash_column, get_rainbow_table_features_column


class FeaturesFilteringManager:
    def get_latest(
        self,
        feature_store: DataFrame,
        rainbow_table: DataFrame,
        features: List[str],
    ):
        if not features:
            features = self.__get_registered_features(feature_store)

        self.__check_features_exist(feature_store, features)

        id_column = feature_store.columns[0]
        time_column = feature_store.columns[1]

        features_data = feature_store.join(rainbow_table, on=get_rainbow_table_hash_column().name).withColumn(
            "features_intersect", f.array_intersect(get_rainbow_table_features_column().name, f.array(*map(f.lit, features)))
        )

        relevant_features = features_data.select(f.array_distinct(f.flatten(f.collect_set("features_intersect")))).collect()[0][0]

        features_data = (
            features_data.groupBy(id_column)
            .agg(
                *[
                    f.array_max(
                        f.collect_list(f.struct(f.array_contains("features_intersect", feature), f.col(time_column), f.col(feature)))
                    )[feature].alias(feature)
                    for feature in relevant_features
                ],
                f.array_distinct(f.flatten(f.collect_set("features_intersect"))).alias("computed_features"),
            )
            .filter(f.size("computed_features") == len(relevant_features))
            .drop("computed_features")
        )

        return features_data

    def get_for_target(
        self,
        feature_store: DataFrame,
        rainbow_table: DataFrame,
        targets: DataFrame,
        features: List[str],
        skip_incomplete_rows: bool,
    ):
        if not features:
            features = self.__get_registered_features(feature_store)

        self.__check_features_exist(feature_store, features)

        if len(targets.columns) != 2:
            raise Exception("Targets dataframe must have exactly two columns [id, date]")

        features_id_column = feature_store.schema[0]
        features_time_column = feature_store.schema[1]
        targets_id_column = targets.schema[0]
        targets_time_column = targets.schema[1]
        join_columns = [features_id_column.name, features_time_column.name]

        if features_id_column.dataType != targets_id_column.dataType:
            raise Exception(
                f"Features and Targets id column type mismatch "
                f"{features_id_column.dataType.typeName()} != {targets_id_column.dataType.typeName()}"
            )

        if features_time_column.dataType != targets_time_column.dataType:
            raise Exception(
                f"Features and Targets id column type mismatch "
                f"{features_time_column.dataType.typeName()} != {targets_time_column.dataType.typeName()}"
            )

        targets = targets.withColumnRenamed(targets_id_column.name, features_id_column.name).withColumnRenamed(
            targets_time_column.name, features_time_column.name
        )

        features_data = (
            feature_store.join(targets, on=join_columns)
            .join(rainbow_table, on=get_rainbow_table_hash_column().name)
            .withColumn("features_intersect", f.array_intersect(get_rainbow_table_features_column().name, f.array(*map(f.lit, features))))
            .withColumn("is_complete_row", f.size("features_intersect") == len(features))
        )

        has_incomplete_rows = len(features_data.filter(~f.col("is_complete_row")).limit(1).collect()) == 1

        if has_incomplete_rows and not skip_incomplete_rows:
            raise Exception("Features contain incomplete rows")

        return features_data.filter(f.col("is_complete_row")).select(*join_columns, *features)

    def __check_features_exist(self, feature_store: DataFrame, features: List[str]):
        registered_features_list = self.__get_registered_features(feature_store)
        unregistered_features = set(features) - set(registered_features_list)

        if unregistered_features != set():
            raise Exception(f"Features {','.join(unregistered_features)} not registered")

    def __get_registered_features(self, feature_store: DataFrame):  # noqa
        return list(feature_store.columns[3:])



import unittest
import hashlib
import datetime as dt
from pyspark.sql import types as t
from pyfonycore.bootstrap import bootstrapped_container
from featurestorebundle.entity.Entity import Entity
from featurestorebundle.delta.feature.FeaturesFilteringManager import FeaturesFilteringManager
from featurestorebundle.test.PySparkTestCase import PySparkTestCase


class FeaturesFilteringManagerTargetsTest(PySparkTestCase):
    def setUp(self) -> None:
        self.__entity = Entity(
            name="client_test",
            id_column="client_id",
            id_column_type=t.StringType(),
            time_column="run_date",
            time_column_type=t.DateType(),
        )

        self.__container = bootstrapped_container.init("test")
        self.__filtering_manager: FeaturesFilteringManager = self.__container.get(FeaturesFilteringManager)

    def test_simple(self):
        feature_store = self.spark.createDataFrame(
            [
                ["1", dt.date(2020, 1, 1), hashlib.md5("f1`f2".encode()).hexdigest(), "c1f1", "c1f2"],
                ["2", dt.date(2018, 1, 1), hashlib.md5("f1`f2".encode()).hexdigest(), "c2f1", "c2f2"],
            ],
            [self.__entity.id_column, self.__entity.time_column, "features_hash", "f1", "f2"],
        )

        rainbow_table = self.spark.createDataFrame(
            [
                [hashlib.md5("f1`f2".encode()).hexdigest(), ["f1", "f2"]],
            ],
            ["features_hash", "computed_columns"],
        )

        targets_table = self.spark.createDataFrame(
            [
                ["1", dt.date(2020, 1, 1)],
                ["2", dt.date(2018, 1, 1)],
            ],
            [self.__entity.id_column, "target_date"],
        )

        features_data = self.__filtering_manager.get_for_target(feature_store, rainbow_table, targets_table, ["f1", "f2"], False)

        expected_features_data = self.spark.createDataFrame(
            [
                ["1", dt.date(2020, 1, 1), "c1f1", "c1f2"],
                ["2", dt.date(2018, 1, 1), "c2f1", "c2f2"],
            ],
            [self.__entity.id_column, self.__entity.time_column, "f1", "f2"],
        )

        self.compare_dataframes(features_data, expected_features_data, self.__entity.id_column)

    def test_incomplete_rows_raise_exception(self):
        feature_store = self.spark.createDataFrame(
            [
                ["1", dt.date(2020, 1, 1), hashlib.md5("f1`f2".encode()).hexdigest(), "c1f1", "c1f2"],
                ["2", dt.date(2018, 1, 1), hashlib.md5("f1".encode()).hexdigest(), "c2f1", None],
            ],
            [self.__entity.id_column, self.__entity.time_column, "features_hash", "f1", "f2"],
        )

        rainbow_table = self.spark.createDataFrame(
            [
                [hashlib.md5("f1`f2".encode()).hexdigest(), ["f1", "f2"]],
                [hashlib.md5("f1".encode()).hexdigest(), ["f1"]],
            ],
            ["features_hash", "computed_columns"],
        )

        targets_table = self.spark.createDataFrame(
            [
                ["1", dt.date(2020, 1, 1)],
                ["2", dt.date(2018, 1, 1)],
            ],
            [self.__entity.id_column, "target_date"],
        )

        with self.assertRaisesRegex(Exception, "Features contain incomplete rows"):
            self.__filtering_manager.get_for_target(feature_store, rainbow_table, targets_table, ["f1", "f2"], False)

    def test_skip_incomplete_rows(self):
        feature_store = self.spark.createDataFrame(
            [
                ["1", dt.date(2020, 1, 1), hashlib.md5("f1`f2".encode()).hexdigest(), "c1f1", "c1f2"],
                ["2", dt.date(2018, 1, 1), hashlib.md5("f1".encode()).hexdigest(), "c2f1", None],
            ],
            [self.__entity.id_column, self.__entity.time_column, "features_hash", "f1", "f2"],
        )

        rainbow_table = self.spark.createDataFrame(
            [
                [hashlib.md5("f1`f2".encode()).hexdigest(), ["f1", "f2"]],
                [hashlib.md5("f1".encode()).hexdigest(), ["f1"]],
            ],
            ["features_hash", "computed_columns"],
        )

        targets_table = self.spark.createDataFrame(
            [
                ["1", dt.date(2020, 1, 1)],
                ["2", dt.date(2018, 1, 1)],
            ],
            [self.__entity.id_column, "target_date"],
        )

        features_data = self.__filtering_manager.get_for_target(feature_store, rainbow_table, targets_table, ["f1", "f2"], True)

        expected_features_data = self.spark.createDataFrame(
            [
                ["1", dt.date(2020, 1, 1), "c1f1", "c1f2"],
            ],
            [self.__entity.id_column, self.__entity.time_column, "f1", "f2"],
        )

        self.compare_dataframes(features_data, expected_features_data, self.__entity.id_column)


if __name__ == "__main__":
    unittest.main()



from typing import Union, Optional
from logging import Logger
from pyspark.sql.types import StructType
from featurestorebundle.delta.PathExistenceChecker import PathExistenceChecker
from featurestorebundle.delta.EmptyDataFrameCreator import EmptyDataFrameCreator


class PathCreator:
    def __init__(
        self,
        logger: Logger,
        path_existence_checker: PathExistenceChecker,
        empty_dataframe_creator: EmptyDataFrameCreator,
    ):
        self.__logger = logger
        self.__path_existence_checker = path_existence_checker
        self.__empty_dataframe_creator = empty_dataframe_creator

    def create_if_not_exists(self, path: str, schema: StructType, partition_by: Optional[Union[str, list]] = None):
        partition_by = partition_by or []
        partition_by = [partition_by] if isinstance(partition_by, str) else partition_by

        if self.__path_existence_checker.exists(path):
            self.__logger.info(f"Delta table at path {path} already exists, creation skipped")
            return

        self.__logger.info(f"Creating new delta table at path {path}")

        df = self.__empty_dataframe_creator.create(schema)

        df.write.partitionBy(*partition_by).format("delta").save(path)

        self.__logger.info("Delta table successfully created")



import unittest
import hashlib
import datetime as dt
from pyspark.sql import types as t
from pyfonycore.bootstrap import bootstrapped_container
from featurestorebundle.entity.Entity import Entity
from featurestorebundle.feature.Feature import Feature
from featurestorebundle.feature.FeatureList import FeatureList
from featurestorebundle.feature.FeatureTemplate import FeatureTemplate
from featurestorebundle.delta.feature.FeaturesPreparer import FeaturesPreparer
from featurestorebundle.feature.FeaturesStorage import FeaturesStorage
from featurestorebundle.delta.feature.schema import get_feature_store_initial_schema, get_rainbow_table_schema
from featurestorebundle.test.PySparkTestCase import PySparkTestCase


class FeaturesPreparerTest(PySparkTestCase):
    def setUp(self) -> None:
        self.__entity = Entity(
            name="client_test",
            id_column="client_id",
            id_column_type=t.StringType(),
            time_column="run_date",
            time_column_type=t.DateType(),
        )

        self.__container = bootstrapped_container.init("test")
        self.__features_preparer: FeaturesPreparer = self.__container.get(FeaturesPreparer)
        self.__feature_store_merge_columns = [self.__entity.id_column, self.__entity.time_column]
        self.__rainbow_table_merge_columns = ["features_hash"]

    def test_simple(self):
        features_storage = FeaturesStorage(self.__entity)
        feature_store = self.spark.createDataFrame([], get_feature_store_initial_schema(self.__entity))
        rainbow_table = self.spark.createDataFrame([], get_rainbow_table_schema())

        df_1 = self.spark.createDataFrame(
            [
                ["1", dt.date(2020, 1, 1), "c1f1", "c1f2"],
                ["2", dt.date(2020, 1, 1), "c2f1", "c2f2"],
            ],
            [self.__entity.id_column, self.__entity.time_column, "f1", "f2"],
        )

        feature_list_1 = FeatureList(
            [
                Feature(self.__entity.name, "f1", "this is feature 1", "string", {}, FeatureTemplate("f1", "this is feature 1")),
                Feature(self.__entity.name, "f2", "this is feature 2", "string", {}, FeatureTemplate("f2", "this is feature 2")),
            ]
        )

        features_storage.add(df_1, feature_list_1)

        features_data, rainbow_data = self.__features_preparer.prepare(features_storage, feature_store, rainbow_table)

        expected_features_data = self.spark.createDataFrame(
            [
                ["1", dt.date(2020, 1, 1), hashlib.md5("f1`f2".encode()).hexdigest(), "c1f1", "c1f2"],
                ["2", dt.date(2020, 1, 1), hashlib.md5("f1`f2".encode()).hexdigest(), "c2f1", "c2f2"],
            ],
            [self.__entity.id_column, self.__entity.time_column, "features_hash", "f1", "f2"],
        )

        expected_rainbow_data = self.spark.createDataFrame(
            [
                [hashlib.md5("f1`f2".encode()).hexdigest(), ["f1", "f2"]],
            ],
            ["features_hash", "computed_columns"],
        )

        self.compare_dataframes(features_data, expected_features_data, self.__feature_store_merge_columns)
        self.compare_dataframes(rainbow_data, expected_rainbow_data, self.__rainbow_table_merge_columns)

    def test_two_feature_results(self):
        features_storage = FeaturesStorage(self.__entity)
        feature_store = self.spark.createDataFrame([], get_feature_store_initial_schema(self.__entity))
        rainbow_table = self.spark.createDataFrame([], get_rainbow_table_schema())

        df_1 = self.spark.createDataFrame(
            [
                ["1", dt.date(2020, 1, 1), "c1f1", "c1f2"],
                ["2", dt.date(2020, 1, 1), "c2f1", "c2f2"],
            ],
            [self.__entity.id_column, self.__entity.time_column, "f1", "f2"],
        )

        feature_list_1 = FeatureList(
            [
                Feature(self.__entity.name, "f1", "this is feature 1", "string", {}, FeatureTemplate("f1", "this is feature 1")),
                Feature(self.__entity.name, "f2", "this is feature 2", "string", {}, FeatureTemplate("f2", "this is feature 2")),
            ]
        )

        df_2 = self.spark.createDataFrame(
            [
                ["1", dt.date(2020, 1, 1), "c1f3"],
                ["2", dt.date(2020, 1, 1), "c2f3"],
                ["3", dt.date(2020, 1, 1), "c3f3"],
            ],
            [self.__entity.id_column, self.__entity.time_column, "f3"],
        )

        feature_list_2 = FeatureList(
            [
                Feature(self.__entity.name, "f3", "this is feature 3", "string", {}, FeatureTemplate("f3", "this is feature 3")),
            ]
        )

        features_storage.add(df_1, feature_list_1)
        features_storage.add(df_2, feature_list_2)

        features_data, rainbow_data = self.__features_preparer.prepare(features_storage, feature_store, rainbow_table)

        expected_features_data = self.spark.createDataFrame(
            [
                ["1", dt.date(2020, 1, 1), hashlib.md5("f1`f2`f3".encode()).hexdigest(), "c1f1", "c1f2", "c1f3"],
                ["2", dt.date(2020, 1, 1), hashlib.md5("f1`f2`f3".encode()).hexdigest(), "c2f1", "c2f2", "c2f3"],
                ["3", dt.date(2020, 1, 1), hashlib.md5("f1`f2`f3".encode()).hexdigest(), None, None, "c3f3"],
            ],
            [self.__entity.id_column, self.__entity.time_column, "features_hash", "f1", "f2", "f3"],
        )

        expected_rainbow_data = self.spark.createDataFrame(
            [
                [hashlib.md5("f1`f2`f3".encode()).hexdigest(), ["f1", "f2", "f3"]],
            ],
            ["features_hash", "computed_columns"],
        )

        self.compare_dataframes(features_data, expected_features_data, self.__feature_store_merge_columns)
        self.compare_dataframes(rainbow_data, expected_rainbow_data, self.__rainbow_table_merge_columns)

    def test_add_new_feature(self):
        features_storage = FeaturesStorage(self.__entity)

        feature_store = self.spark.createDataFrame(
            [
                ["1", dt.date(2020, 1, 1), hashlib.md5("f1`f2".encode()).hexdigest(), "c1f1", "c1f2"],
                ["2", dt.date(2020, 1, 1), hashlib.md5("f1`f2".encode()).hexdigest(), "c2f1", "c2f2"],
            ],
            [self.__entity.id_column, self.__entity.time_column, "features_hash", "f1", "f2"],
        )

        rainbow_table = self.spark.createDataFrame(
            [
                [hashlib.md5("f1`f2".encode()).hexdigest(), ["f1", "f2"]],
            ],
            ["features_hash", "computed_columns"],
        )

        df_1 = self.spark.createDataFrame(
            [
                ["1", dt.date(2020, 1, 2), "c1f3"],
            ],
            [self.__entity.id_column, self.__entity.time_column, "f3"],
        )

        feature_list_1 = FeatureList(
            [
                Feature(self.__entity.name, "f3", "this is feature 3", "string", {}, FeatureTemplate("f3", "this is feature 3")),
            ]
        )

        features_storage.add(df_1, feature_list_1)

        features_data, rainbow_data = self.__features_preparer.prepare(features_storage, feature_store, rainbow_table)

        feature_store_after_merge = self.delta_merge(feature_store, features_data, self.__feature_store_merge_columns)
        rainbow_table_after_merge = self.delta_merge(rainbow_table, rainbow_data, self.__rainbow_table_merge_columns)

        expected_feature_store = self.spark.createDataFrame(
            [
                ["1", dt.date(2020, 1, 1), hashlib.md5("f1`f2".encode()).hexdigest(), "c1f1", "c1f2", None],
                ["2", dt.date(2020, 1, 1), hashlib.md5("f1`f2".encode()).hexdigest(), "c2f1", "c2f2", None],
                ["1", dt.date(2020, 1, 2), hashlib.md5("f3".encode()).hexdigest(), None, None, "c1f3"],
            ],
            [self.__entity.id_column, self.__entity.time_column, "features_hash", "f1", "f2", "f3"],
        )

        expected_rainbow_table = self.spark.createDataFrame(
            [
                [hashlib.md5("f1`f2".encode()).hexdigest(), ["f1", "f2"]],
                [hashlib.md5("f3".encode()).hexdigest(), ["f3"]],
            ],
            ["features_hash", "computed_columns"],
        )

        self.compare_dataframes(feature_store_after_merge, expected_feature_store, self.__feature_store_merge_columns)
        self.compare_dataframes(rainbow_table_after_merge, expected_rainbow_table, self.__rainbow_table_merge_columns)

    def test_backfill_feature(self):
        features_storage = FeaturesStorage(self.__entity)

        feature_store = self.spark.createDataFrame(
            [
                ["1", dt.date(2020, 1, 1), hashlib.md5("f1`f2".encode()).hexdigest(), "c1f1", "c1f2", None],
                ["2", dt.date(2020, 1, 1), hashlib.md5("f1`f2".encode()).hexdigest(), "c2f1", "c2f2", None],
                ["1", dt.date(2020, 1, 2), hashlib.md5("f3".encode()).hexdigest(), None, None, "c1f3"],
            ],
            [self.__entity.id_column, self.__entity.time_column, "features_hash", "f1", "f2", "f3"],
        )

        rainbow_table = self.spark.createDataFrame(
            [
                [hashlib.md5("f1`f2".encode()).hexdigest(), ["f1", "f2"]],
                [hashlib.md5("f3".encode()).hexdigest(), ["f3"]],
            ],
            ["features_hash", "computed_columns"],
        )

        df_1 = self.spark.createDataFrame(
            [
                ["1", dt.date(2020, 1, 1), "c1f3"],
            ],
            [self.__entity.id_column, self.__entity.time_column, "f3"],
        )

        feature_list_1 = FeatureList(
            [
                Feature(self.__entity.name, "f3", "this is feature 3", "string", {}, FeatureTemplate("f3", "this is feature 3")),
            ]
        )

        features_storage.add(df_1, feature_list_1)

        features_data, rainbow_data = self.__features_preparer.prepare(features_storage, feature_store, rainbow_table)

        feature_store_after_merge = self.delta_merge(feature_store, features_data, self.__feature_store_merge_columns)
        rainbow_table_after_merge = self.delta_merge(rainbow_table, rainbow_data, self.__rainbow_table_merge_columns)

        expected_feature_store = self.spark.createDataFrame(
            [
                ["1", dt.date(2020, 1, 1), hashlib.md5("f1`f2`f3".encode()).hexdigest(), "c1f1", "c1f2", "c1f3"],
                ["2", dt.date(2020, 1, 1), hashlib.md5("f1`f2".encode()).hexdigest(), "c2f1", "c2f2", None],
                ["1", dt.date(2020, 1, 2), hashlib.md5("f3".encode()).hexdigest(), None, None, "c1f3"],
            ],
            [self.__entity.id_column, self.__entity.time_column, "features_hash", "f1", "f2", "f3"],
        )

        expected_rainbow_table = self.spark.createDataFrame(
            [
                [hashlib.md5("f1`f2`f3".encode()).hexdigest(), ["f1", "f2", "f3"]],
                [hashlib.md5("f1`f2".encode()).hexdigest(), ["f1", "f2"]],
                [hashlib.md5("f3".encode()).hexdigest(), ["f3"]],
            ],
            ["features_hash", "computed_columns"],
        )

        self.compare_dataframes(feature_store_after_merge, expected_feature_store, self.__feature_store_merge_columns)
        self.compare_dataframes(rainbow_table_after_merge, expected_rainbow_table, self.__rainbow_table_merge_columns)

    def test_overwrite_feature(self):
        features_storage = FeaturesStorage(self.__entity)

        feature_store = self.spark.createDataFrame(
            [
                ["1", dt.date(2020, 1, 1), hashlib.md5("f1`f2`f3".encode()).hexdigest(), "c1f1", "c1f2", "c1f2"],
                ["2", dt.date(2020, 1, 1), hashlib.md5("f1`f2`f3".encode()).hexdigest(), "c2f1", "c2f2", "c2f3"],
                ["1", dt.date(2020, 1, 2), hashlib.md5("f1`f2`f3".encode()).hexdigest(), "c1f1", "c1f2", "c1f3"],
            ],
            [self.__entity.id_column, self.__entity.time_column, "features_hash", "f1", "f2", "f3"],
        )

        rainbow_table = self.spark.createDataFrame(
            [
                [hashlib.md5("f1`f2`f3".encode()).hexdigest(), ["f1", "f2", "f3"]],
            ],
            ["features_hash", "computed_columns"],
        )

        df_1 = self.spark.createDataFrame(
            [
                ["1", dt.date(2020, 1, 1), "xxx"],
            ],
            [self.__entity.id_column, self.__entity.time_column, "f3"],
        )

        feature_list_1 = FeatureList(
            [
                Feature(self.__entity.name, "f3", "this is feature 3", "string", {}, FeatureTemplate("f3", "this is feature 3")),
            ]
        )

        df_2 = self.spark.createDataFrame(
            [
                ["1", dt.date(2020, 1, 2), "xxx"],
            ],
            [self.__entity.id_column, self.__entity.time_column, "f1"],
        )

        feature_list_2 = FeatureList(
            [
                Feature(self.__entity.name, "f1", "this is feature 1", "string", {}, FeatureTemplate("f1", "this is feature 1")),
            ]
        )

        features_storage.add(df_1, feature_list_1)
        features_storage.add(df_2, feature_list_2)

        features_data, rainbow_data = self.__features_preparer.prepare(features_storage, feature_store, rainbow_table)

        feature_store_after_merge = self.delta_merge(feature_store, features_data, self.__feature_store_merge_columns)
        rainbow_table_after_merge = self.delta_merge(rainbow_table, rainbow_data, self.__rainbow_table_merge_columns)

        expected_feature_store = self.spark.createDataFrame(
            [
                ["1", dt.date(2020, 1, 1), hashlib.md5("f1`f2`f3".encode()).hexdigest(), "c1f1", "c1f2", "xxx"],
                ["2", dt.date(2020, 1, 1), hashlib.md5("f1`f2`f3".encode()).hexdigest(), "c2f1", "c2f2", "c2f3"],
                ["1", dt.date(2020, 1, 2), hashlib.md5("f1`f2`f3".encode()).hexdigest(), "xxx", "c1f2", "c1f3"],
            ],
            [self.__entity.id_column, self.__entity.time_column, "features_hash", "f1", "f2", "f3"],
        )

        expected_rainbow_table = self.spark.createDataFrame(
            [
                [hashlib.md5("f1`f2`f3".encode()).hexdigest(), ["f1", "f2", "f3"]],
            ],
            ["features_hash", "computed_columns"],
        )

        self.compare_dataframes(feature_store_after_merge, expected_feature_store, self.__feature_store_merge_columns)
        self.compare_dataframes(rainbow_table_after_merge, expected_rainbow_table, self.__rainbow_table_merge_columns)

    def test_dynamic_run_date(self):
        features_storage = FeaturesStorage(self.__entity)
        feature_store = self.spark.createDataFrame([], get_feature_store_initial_schema(self.__entity))
        rainbow_table = self.spark.createDataFrame([], get_rainbow_table_schema())

        df_1 = self.spark.createDataFrame(
            [
                ["1", dt.date(2020, 1, 1), "c1f1", "c1f2"],
                ["2", dt.date(2020, 1, 1), "c2f1", "c2f2"],
            ],
            [self.__entity.id_column, self.__entity.time_column, "f1", "f2"],
        )

        feature_list_1 = FeatureList(
            [
                Feature(self.__entity.name, "f1", "this is feature 1", "string", {}, FeatureTemplate("f1", "this is feature 1")),
                Feature(self.__entity.name, "f2", "this is feature 2", "string", {}, FeatureTemplate("f2", "this is feature 2")),
            ]
        )

        df_2 = self.spark.createDataFrame(
            [
                ["1", dt.date(2020, 1, 2), "c1f3"],
            ],
            [self.__entity.id_column, self.__entity.time_column, "f3"],
        )

        feature_list_2 = FeatureList(
            [
                Feature(self.__entity.name, "f3", "this is feature 3", "string", {}, FeatureTemplate("f3", "this is feature 3")),
            ]
        )

        features_storage.add(df_1, feature_list_1)
        features_storage.add(df_2, feature_list_2)

        features_data, rainbow_data = self.__features_preparer.prepare(features_storage, feature_store, rainbow_table)

        expected_features_data = self.spark.createDataFrame(
            [
                ["1", dt.date(2020, 1, 1), hashlib.md5("f1`f2`f3".encode()).hexdigest(), "c1f1", "c1f2", None],
                ["2", dt.date(2020, 1, 1), hashlib.md5("f1`f2`f3".encode()).hexdigest(), "c2f1", "c2f2", None],
                ["1", dt.date(2020, 1, 2), hashlib.md5("f1`f2`f3".encode()).hexdigest(), None, None, "c1f3"],
            ],
            [self.__entity.id_column, self.__entity.time_column, "features_hash", "f1", "f2", "f3"],
        )

        expected_rainbow_data = self.spark.createDataFrame(
            [
                [hashlib.md5("f1`f2`f3".encode()).hexdigest(), ["f1", "f2", "f3"]],
            ],
            ["features_hash", "computed_columns"],
        )

        self.compare_dataframes(features_data, expected_features_data, self.__feature_store_merge_columns)
        self.compare_dataframes(rainbow_data, expected_rainbow_data, self.__rainbow_table_merge_columns)

    def test_more_complicated(self):
        features_storage = FeaturesStorage(self.__entity)
        feature_store = self.spark.createDataFrame([], get_feature_store_initial_schema(self.__entity))
        rainbow_table = self.spark.createDataFrame([], get_rainbow_table_schema())

        df_1 = self.spark.createDataFrame(
            [
                ["1", dt.date(2020, 1, 1), "c1f1", "c1f2"],
                ["2", dt.date(2020, 1, 2), "c2f1", "c2f2"],
            ],
            [self.__entity.id_column, self.__entity.time_column, "f1", "f2"],
        )

        feature_list_1 = FeatureList(
            [
                Feature(self.__entity.name, "f1", "this is feature 1", "string", {}, FeatureTemplate("f1", "this is feature 1")),
                Feature(self.__entity.name, "f2", "this is feature 2", "string", {}, FeatureTemplate("f2", "this is feature 2")),
            ]
        )

        df_2 = self.spark.createDataFrame(
            [
                ["1", dt.date(2020, 1, 3), "c1f3"],
                ["2", dt.date(2020, 1, 2), "c2f3"],
                ["3", dt.date(2020, 1, 3), "c3f3"],
            ],
            [self.__entity.id_column, self.__entity.time_column, "f3"],
        )

        feature_list_2 = FeatureList(
            [
                Feature(self.__entity.name, "f3", "this is feature 3", "string", {}, FeatureTemplate("f3", "this is feature 3")),
            ]
        )

        features_storage.add(df_1, feature_list_1)
        features_storage.add(df_2, feature_list_2)

        features_data, rainbow_data = self.__features_preparer.prepare(features_storage, feature_store, rainbow_table)

        expected_features_data = self.spark.createDataFrame(
            [
                ["1", dt.date(2020, 1, 1), hashlib.md5("f1`f2`f3".encode()).hexdigest(), "c1f1", "c1f2", None],
                ["2", dt.date(2020, 1, 2), hashlib.md5("f1`f2`f3".encode()).hexdigest(), "c2f1", "c2f2", "c2f3"],
                ["1", dt.date(2020, 1, 3), hashlib.md5("f1`f2`f3".encode()).hexdigest(), None, None, "c1f3"],
                ["3", dt.date(2020, 1, 3), hashlib.md5("f1`f2`f3".encode()).hexdigest(), None, None, "c3f3"],
            ],
            [self.__entity.id_column, self.__entity.time_column, "features_hash", "f1", "f2", "f3"],
        )

        expected_rainbow_data = self.spark.createDataFrame(
            [
                [hashlib.md5("f1`f2`f3".encode()).hexdigest(), ["f1", "f2", "f3"]],
            ],
            ["features_hash", "computed_columns"],
        )

        self.compare_dataframes(features_data, expected_features_data, self.__feature_store_merge_columns)
        self.compare_dataframes(rainbow_data, expected_rainbow_data, self.__rainbow_table_merge_columns)


if __name__ == "__main__":
    unittest.main()



from typing import List, Set

from pyspark.sql import DataFrame

from featurestorebundle.entity.Entity import Entity
from featurestorebundle.feature.Feature import Feature
from featurestorebundle.feature.FeatureList import FeatureList
from featurestorebundle.feature.FeaturePattern import FeaturePattern
from featurestorebundle.feature.FeatureTemplate import FeatureTemplate
from featurestorebundle.metadata.TimeWindowHandler import TimeWindowHandler


class TemplateMatchingError(Exception):
    pass


class FeatureTemplateMatcher:
    def __init__(self, time_window_handler: TimeWindowHandler):
        self.__time_window_handler = time_window_handler

    def get_features(self, entity: Entity, feature_templates: List[FeatureTemplate], df: DataFrame) -> FeatureList:
        pk_columns = [entity.id_column, entity.time_column]

        feature_columns = [col for col in df.schema.jsonValue()["fields"] if col["name"] not in pk_columns]
        feature_patterns = [FeaturePattern(feature_template) for feature_template in feature_templates]
        unmatched_patterns = set(feature_patterns)

        features = [
            self.__get_feature(entity.name, col["name"], col["type"], feature_patterns, unmatched_patterns) for col in feature_columns
        ]

        if unmatched_patterns:
            patterns = ", ".join(f'"{pattern.feature_template}"' for pattern in unmatched_patterns)
            raise TemplateMatchingError(f"Templates {patterns} did not match any columns.")

        return FeatureList(features)

    def __get_feature(
        self, entity: str, name: str, dtype: str, feature_patterns: List[FeaturePattern], unmatched_patterns: Set[FeaturePattern]
    ) -> Feature:
        for feature_pattern in feature_patterns:
            feature_template = feature_pattern.feature_template
            match = feature_pattern.get_match(name)

            if not match:
                continue

            unmatched_patterns.discard(feature_pattern)
            metadata = feature_pattern.get_groups_as_dict(match)

            if "time_window" in metadata:
                time_window = metadata["time_window"]
                self.__time_window_handler.is_valid(time_window, name, feature_pattern)

            description = feature_template.description_template.format(
                **{
                    key: (self.__time_window_handler.to_text(val) if key == "time_window" else val.replace("_", " "))
                    for key, val in metadata.items()
                }
            )

            return Feature(entity, name, description, dtype, metadata, feature_template)

        raise TemplateMatchingError(f"Column '{name}' could not be matched by any template.")



from typing import Dict, List, Union

from featurestorebundle.feature.FeatureTemplate import FeatureTemplate


class Feature:
    def __init__(self, entity: str, name: str, description: str, dtype: str, extra: Dict, template: FeatureTemplate):
        self.__entity = entity
        self.__name = name
        self.__description = description
        self.__dtype = dtype
        self.__extra = extra
        self.__template = template

    @property
    def entity(self):
        return self.__entity

    @property
    def name(self):
        return self.__name

    @property
    def description(self):
        return self.__description

    @property
    def dtype(self):
        return self.__dtype

    def get_metadata_dict(self) -> Dict[str, Union[str, Dict[str, str]]]:
        return {
            "entity": self.__entity,
            "name": self.__name,
            "description": self.__description,
            "extra": self.__extra,
            "template": self.__template.name_template,
            "category": self.__template.category,
            "dtype": self.__dtype,
        }

    def get_metadata_list(self) -> List[Union[Dict[str, str], str]]:
        return list(self.get_metadata_dict().values())



from featurestorebundle.db.TableNames import TableNames
from featurestorebundle.feature.FeaturesStorage import FeaturesStorage
from featurestorebundle.delta.feature.FeaturesPreparer import FeaturesPreparer
from featurestorebundle.feature.FeaturesValidator import FeaturesValidator
from featurestorebundle.delta.feature.DeltaRainbowTableManager import DeltaRainbowTableManager
from featurestorebundle.delta.feature.writer.DeltaFeaturesDataHandler import DeltaFeaturesDataHandler
from featurestorebundle.delta.feature.writer.DeltaFeaturesMergeConfigGenerator import DeltaFeaturesMergeConfigGenerator
from featurestorebundle.delta.feature.writer.DeltaTableFeaturesPreparer import DeltaTableFeaturesPreparer
from featurestorebundle.feature.reader.FeaturesReaderInterface import FeaturesReaderInterface
from featurestorebundle.feature.writer.FeaturesWriterInterface import FeaturesWriterInterface
from featurestorebundle.metadata.writer.MetadataWriterInterface import MetadataWriterInterface


# pylint: disable=too-many-instance-attributes
class DeltaTableFeaturesWriter(FeaturesWriterInterface):
    def __init__(
        self,
        features_reader: FeaturesReaderInterface,
        metadata_writer: MetadataWriterInterface,
        delta_data_handler: DeltaFeaturesDataHandler,
        delta_merge_config_generator: DeltaFeaturesMergeConfigGenerator,
        features_table_preparer: DeltaTableFeaturesPreparer,
        rainbow_table_manager: DeltaRainbowTableManager,
        features_preparer: FeaturesPreparer,
        features_validator: FeaturesValidator,
        table_names: TableNames,
    ):
        self.__features_reader = features_reader
        self.__metadata_writer = metadata_writer
        self.__delta_data_handler = delta_data_handler
        self.__delta_merge_config_generator = delta_merge_config_generator
        self.__features_table_preparer = features_table_preparer
        self.__rainbow_table_manager = rainbow_table_manager
        self.__features_preparer = features_preparer
        self.__features_validator = features_validator
        self.__table_names = table_names

    def write(self, features_storage: FeaturesStorage):
        entity = features_storage.entity
        feature_list = features_storage.feature_list
        full_table_name = self.__table_names.get_features_full_table_name(entity.name)
        path = self.__table_names.get_features_path(entity.name)
        pk_columns = [entity.id_column, entity.time_column]

        feature_store = self.__features_reader.read_safe(entity)
        rainbow_table = self.__rainbow_table_manager.read_safe(entity.name)
        features_data, rainbow_data = self.__features_preparer.prepare(features_storage, feature_store, rainbow_table)
        delta_merge_config = self.__delta_merge_config_generator.generate(entity, features_data, pk_columns)

        self.__features_validator.validate(entity, features_data, feature_list)
        self.__features_table_preparer.prepare(full_table_name, path, entity, feature_list)
        self.__rainbow_table_manager.merge(entity.name, rainbow_data)
        self.__delta_data_handler.merge_to_delta_table(full_table_name, delta_merge_config)
        self.__metadata_writer.write(entity, feature_list)



from abc import ABC, abstractmethod
from featurestorebundle.feature.FeaturesStorage import FeaturesStorage


class FeaturesWriterInterface(ABC):
    @abstractmethod
    def write(self, features_storage: FeaturesStorage):
        pass



from pyspark.sql import SparkSession


class TableExistenceChecker:
    def __init__(self, spark: SparkSession):
        self.__spark = spark

    def exists(self, full_table_name: str) -> bool:
        db_name = full_table_name.split(".")[0]
        table_name = full_table_name.split(".")[1]
        databases = [db.databaseName for db in self.__spark.sql("SHOW DATABASES").collect()]

        if db_name not in databases:
            return False

        return self.__spark.sql(f'SHOW TABLES IN {db_name} LIKE "{table_name}"').collect() != []



import os
import pyspark.sql.types as t
import unittest
from collections import namedtuple
from featurestorebundle.delta.feature.writer.DeltaFeaturesMergeConfigGenerator import DeltaFeaturesMergeConfigGenerator
from featurestorebundle.entity.Entity import Entity

os.environ["APP_ENV"] = "test"


class FakeSchema:
    client_id = {"name": "client_id", "type": "long"}
    run_date = {"name": "run_date", "type": "long"}
    my_sample_feature = {"name": "my_sample_feature", "type": "long"}

    schema = {
        "fields": [
            namedtuple("ObjectName", client_id.keys())(*client_id.values()),
            namedtuple("ObjectName", run_date.keys())(*run_date.values()),
            namedtuple("ObjectName", my_sample_feature.keys())(*my_sample_feature.values()),
        ],
    }

    fields = schema["fields"]

    @staticmethod
    def jsonValue():  # noqa N802 pylint: disable=invalid-name
        return FakeSchema.schema


class FakeResult:

    columns = ["client_id", "run_date", "my_sample_feature"]

    def __init__(self, value):
        self.value = value
        self.schema = FakeSchema


class DeltaFeaturesMergeConfigGeneratorTest(unittest.TestCase):
    def setUp(self) -> None:
        self.entity = Entity(
            name="client_test",
            id_column="client_id",
            id_column_type=t.StringType(),
            time_column="run_date",
            time_column_type=t.DateType(),
        )
        fake_df = FakeResult("not_a_real_dataframe")

        self.__config = DeltaFeaturesMergeConfigGenerator().generate(
            self.entity, fake_df, [self.entity.id_column, self.entity.time_column]  # noqa # pyre-ignore[6]
        )

    def test_run_date_update_set(self):
        self.assertNotIn("run_date", self.__config.update_set)

    def test_run_date_insert_set(self):
        self.assertIn("run_date", self.__config.insert_set)

    def test_client_id_update_set(self):
        self.assertNotIn("client_id", self.__config.update_set)

    def test_client_id_insert_set(self):
        self.assertIn("client_id", self.__config.insert_set)

    def test_feature_in_both_sets(self):
        self.assertIn("my_sample_feature", self.__config.insert_set)

        self.assertIn("my_sample_feature", self.__config.update_set)

    def test_merge_condition(self):
        self.assertEqual(
            " AND ".join(f"target.{pk} = source.{pk}" for pk in [self.entity.id_column, self.entity.time_column]),
            self.__config.merge_condition,
        )


if __name__ == "__main__":
    unittest.main()



from typing import List, Dict, Union
import IPython


class MetadataHTMLDisplayer:
    def display(self, metadata: List[Dict[str, Union[str, Dict[str, str]]]]):
        display_html = self.__get_display_html()
        html = self.__get_html(metadata)
        display_html(html)

    def __get_html(self, metadata: List[Dict[str, Union[str, Dict[str, str]]]]) -> str:
        html = f"""
        <!doctype html>
              <html lang="en">
              <head>
                <meta charset="utf-8">
                <link href="https://cdn.jsdelivr.net/npm/bootstrap@5.1.0/dist/css/bootstrap.min.css" rel="stylesheet" integrity="sha384-KyZXEAg3QhqLMpG8r+8fhAXLRk2vvoC2f3B09zVXn8CA5QIVfZOJ3BCsw2P0p/We" crossorigin="anonymous">
              </head>
              <body>
                <table class="table table-sm small">
                      <thead>
                        <tr>
                          <th scope="col">Name</th>
                          <th scope="col">Description</th>
                          <th scope="col">Extra</th>
                          <th scope="col">Template</th>
                          <th scope="col">Type</th>
                        </tr>
                      </thead>
                      <tbody>
                        {"".join(self.__get_table_row(metadata_dict) for metadata_dict in metadata)}
                      </tbody>
                    </table>
              </body>
            </html>
        """
        return html

    def __get_table_row(self, metadata_dict: Dict[str, Union[str, Dict[str, str]]]) -> str:
        # pyre-fixme[16]:
        return f"""
        <tr>
            <td>{metadata_dict["name"]}</td>
            <td>{metadata_dict["description"]}</td>
            <td>{", ".join(f"<b>{key}</b>: {val}" for key, val in metadata_dict["extra"].items())}</td>
            <td>{metadata_dict["template"]}</td>
            <td>{metadata_dict["dtype"]}</td>
        </tr>
        """

    def __get_display_html(self):
        ipython = IPython.get_ipython()

        if not hasattr(ipython, "user_ns") or "displayHTML" not in ipython.user_ns:
            raise Exception("displayHTML cannot be resolved")

        return ipython.user_ns["displayHTML"]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['featurestorebundle',
 'featurestorebundle.databricks.feature.reader',
 'featurestorebundle.databricks.feature.writer',
 'featurestorebundle.db',
 'featurestorebundle.delta',
 'featurestorebundle.delta.feature',
 'featurestorebundle.delta.feature.reader',
 'featurestorebundle.delta.feature.writer',
 'featurestorebundle.delta.metadata',
 'featurestorebundle.delta.metadata.reader',
 'featurestorebundle.delta.metadata.writer',
 'featurestorebundle.delta.target',
 'featurestorebundle.delta.target.reader',
 'featurestorebundle.entity',
 'featurestorebundle.feature',
 'featurestorebundle.feature.reader',
 'featurestorebundle.feature.writer',
 'featurestorebundle.metadata',
 'featurestorebundle.metadata.reader',
 'featurestorebundle.metadata.writer',
 'featurestorebundle.notebook.decorator',
 'featurestorebundle.notebook.decorator.tests',
 'featurestorebundle.target.reader',
 'featurestorebundle.test',
 'featurestorebundle.widgets',
 'featurestorebundle.windows',
 'featurestorebundle.windows.tests']

package_data = \
{'': ['*'], 'featurestorebundle': ['_config/*']}

install_requires = \
['daipe-core>=1.2,<2.0', 'pyfony-bundles>=0.4.0,<0.5.0']

entry_points = \
{'pyfony.bundle': ['create = '
                   'featurestorebundle.FeatureStoreBundle:FeatureStoreBundle']}

setup_kwargs = {
    'name': 'feature-store-bundle',
    'version': '2.0.0.dev5',
    'description': 'Feature Store for the Daipe AI Platform',
    'long_description': '# Feature Store bundle\n\n**This package is distributed under the "DataSentics SW packages Terms of Use." See [license](https://raw.githubusercontent.com/daipe-ai/feature-store-bundle/master/LICENSE)**\n\nFeature store bundle allows you to store features with metadata.\n\n# Installation\n\n```bash\npoetry add feature-store-bundle\n```\n\n# Getting started\n\n\n1. Define entity and custom `feature decorator`\n\n```python\nfrom pyspark.sql import types as t\nfrom daipecore.decorator.DecoratedDecorator import DecoratedDecorator\nfrom featurestorebundle.entity.Entity import Entity\nfrom featurestorebundle.feature.FeaturesStorage import FeaturesStorage\nfrom featurestorebundle.notebook.decorator.feature import feature\n\nentity = Entity(\n    name="client",\n    id_column="UserName",\n    id_column_type=t.StringType(),\n    time_column="run_date",\n    time_column_type=t.DateType(),\n)\n\n@DecoratedDecorator\nclass client_feature(feature):  # noqa N081\n    def __init__(self, *args, category=None):\n        super().__init__(*args, entity=entity, category=category, features_storage=features_storage)\n```\n\n2. Use the `feature decorator` to save features as you create them\n\n```python\nfrom pyspark.sql import functions as f\nfrom pyspark.sql import DataFrame\nfrom datalakebundle.imports import transformation, read_table\n\n@transformation(read_table("silver.tbl_loans"), display=True)\n@client_feature(\n    ("Age", "Client\'s age"),\n    ("Gender", "Client\'s gender"),\n    ("WorkExperience", "Client\'s work experience"),\n    category="personal",\n)\ndef client_personal_features(df: DataFrame):\n    return (\n        df.select("UserName", "Age", "Gender", "WorkExperience")\n        .groupBy("UserName")\n        .agg(\n            f.max("Age").alias("Age"),\n            f.first("Gender").alias("Gender"),\n            f.first("WorkExperience").alias("WorkExperience"),\n        )\n        .withColumn("run_date", f.lit(today))\n    )\n```\n\n3. Write/Merge all features in one go\n\n```python\nfrom datalakebundle.imports import notebook_function\nfrom featurestorebundle.delta.DeltaWriter import DeltaWriter\n\nnotebook_function()\ndef write_features(writer: DeltaWriter):\n    writer.write_latest(features_storage)\n```\n',
    'author': 'Datasentics',
    'author_email': 'jiri.koutny@datasentics.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/daipe-ai/feature-store-bundle',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from featurestorebundle.feature.FeaturePattern import FeaturePattern

PERIODS = {
    "h": "hours",
    "d": "days",
    "w": "weeks",
}


class TimeWindowFormatError(Exception):
    pass


class TimeWindowHandler:
    def is_valid(self, time_window: str, feature_name: str, feature_pattern: FeaturePattern):
        time_window_value = time_window[:-1]
        time_window_period = time_window[-1]

        if not time_window_value.isdigit():
            raise TimeWindowFormatError(
                f"Column '{feature_name}' has been matched by '{feature_pattern.feature_template.name_template}' and time_window={time_window_value} which is not a positive integer. Check that your templates adhere to the rules at https://docs.daipe.ai/feature-store/templates/"
            )
        if time_window_period not in PERIODS:
            raise TimeWindowFormatError(
                f"Column '{feature_name}' has been matched by '{feature_pattern.feature_template.name_template}' and time_window={time_window} with period '{time_window_period}' is not from supported periods: {', '.join(PERIODS.keys())}"
            )

    def to_text(self, time_window: str) -> str:
        n = int(time_window[:-1])
        period = time_window[-1]

        result = f"{n} {PERIODS[period]}"
        return result[:-1] if n == 1 else result



from typing import List
from pyspark.sql import SparkSession
from featurestorebundle.feature.Feature import Feature
from featurestorebundle.feature.FeatureList import FeatureList


class DeltaFeaturesRegistrator:
    def __init__(self, spark: SparkSession):
        self.__spark = spark

    def register(self, table_identifier: str, feature_list: FeatureList):
        def build_add_column_string(feature: Feature):
            return f'{feature.name} {feature.dtype} COMMENT "{feature.description}"'

        def build_add_columns_string(table_identifier, feature_list: FeatureList):
            add_column_sqls = [build_add_column_string(feature) for feature in feature_list.get_all()]
            return f"ALTER TABLE {table_identifier} ADD COLUMNS ({','.join(add_column_sqls)})"

        registered_feature_names = self.__get_feature_names(table_identifier)
        unregistered_features = feature_list.get_unregistered(registered_feature_names)

        if not unregistered_features.empty():
            self.__spark.sql(build_add_columns_string(table_identifier, feature_list))

    def __get_feature_names(self, table_identifier: str) -> List[str]:
        column_definitions = self.__spark.sql(f"DESCRIBE TABLE {table_identifier}").collect()

        def find_separation_row(column_definitions):
            for i, row in enumerate(column_definitions):
                if row.col_name in ["", "# Partition Information"]:
                    return i

            return None

        feature_rows = column_definitions[3 : find_separation_row(column_definitions)]

        return [row.col_name for row in feature_rows]



from typing import List
from pyspark.sql import DataFrame
from featurestorebundle.entity.Entity import Entity
from featurestorebundle.feature.FeatureList import FeatureList


class FeaturesStorage:
    __entity: Entity
    __results: List[DataFrame]
    __feature_list: FeatureList

    def __init__(self, entity: Entity):
        self.__entity = entity
        self.__results = []
        self.__feature_list = FeatureList([])

    @property
    def entity(self):
        return self.__entity

    @property
    def results(self):
        return self.__results

    @property
    def feature_list(self):
        return self.__feature_list

    def add(self, result: DataFrame, feature_list: FeatureList):
        self.__results.append(result)
        self.__feature_list = self.__feature_list.merge(feature_list)



import unittest
from datetime import datetime
import time
from featurestorebundle.windows.windowed_features import _is_past_time_window


def get_unix_timestamp(date_str: str):
    return time.mktime(datetime.strptime(date_str, "%Y-%m-%d").timetuple())


class WindowedFeaturesTest(unittest.TestCase):
    def test_past_time_windows(self):
        run_date = get_unix_timestamp("2021-11-18")
        arg_date = get_unix_timestamp("2021-10-17")

        self.assertFalse(_is_past_time_window(run_date, arg_date, "30d"))
        self.assertTrue(_is_past_time_window(run_date, arg_date, "60d"))
        self.assertTrue(_is_past_time_window(run_date, arg_date, "90d"))

    def test_wrong_order_time_windows(self):
        arg_date = get_unix_timestamp("2021-11-18")
        run_date = get_unix_timestamp("2021-10-17")

        self.assertFalse(_is_past_time_window(run_date, arg_date, "30d"))
        self.assertFalse(_is_past_time_window(run_date, arg_date, "60d"))
        self.assertFalse(_is_past_time_window(run_date, arg_date, "90d"))


if __name__ == "__main__":
    unittest.main()



import unittest
import hashlib
import datetime as dt
from pyspark.sql import types as t
from pyfonycore.bootstrap import bootstrapped_container
from featurestorebundle.entity.Entity import Entity
from featurestorebundle.delta.feature.FeaturesFilteringManager import FeaturesFilteringManager
from featurestorebundle.test.PySparkTestCase import PySparkTestCase


class FeaturesFilteringManagerLatestTest(PySparkTestCase):
    def setUp(self) -> None:
        self.__entity = Entity(
            name="client_test",
            id_column="client_id",
            id_column_type=t.StringType(),
            time_column="run_date",
            time_column_type=t.DateType(),
        )

        self.__container = bootstrapped_container.init("test")
        self.__filtering_manager: FeaturesFilteringManager = self.__container.get(FeaturesFilteringManager)

    def test_simple(self):
        feature_store = self.spark.createDataFrame(
            [
                ["1", dt.date(2020, 1, 1), hashlib.md5("f1`f2".encode()).hexdigest(), "c1f1", "c1f2"],
                ["2", dt.date(2020, 1, 1), hashlib.md5("f1`f2".encode()).hexdigest(), "c2f1", "c2f2"],
            ],
            [self.__entity.id_column, self.__entity.time_column, "features_hash", "f1", "f2"],
        )

        rainbow_table = self.spark.createDataFrame(
            [
                [hashlib.md5("f1`f2".encode()).hexdigest(), ["f1", "f2"]],
            ],
            ["features_hash", "computed_columns"],
        )

        features_data = self.__filtering_manager.get_latest(feature_store, rainbow_table, ["f1", "f2"])

        expected_features_data = self.spark.createDataFrame(
            [
                ["1", "c1f1", "c1f2"],
                ["2", "c2f1", "c2f2"],
            ],
            [self.__entity.id_column, "f1", "f2"],
        )

        self.compare_dataframes(features_data, expected_features_data, [self.__entity.id_column])

    def test_non_existent_feature_raises_exception(self):
        feature_store = self.spark.createDataFrame(
            [
                ["1", dt.date(2020, 1, 1), hashlib.md5("f1`f2".encode()).hexdigest(), "c1f1", "c1f2"],
                ["2", dt.date(2020, 1, 1), hashlib.md5("f1`f2".encode()).hexdigest(), "c2f1", "c2f2"],
            ],
            [self.__entity.id_column, self.__entity.time_column, "features_hash", "f1", "f2"],
        )

        rainbow_table = self.spark.createDataFrame(
            [
                [hashlib.md5("f1`f2".encode()).hexdigest(), ["f1", "f2"]],
            ],
            ["features_hash", "computed_columns"],
        )

        with self.assertRaisesRegex(Exception, "Features f3 not registered"):
            self.__filtering_manager.get_latest(feature_store, rainbow_table, ["f3"])

    def test_uncomputed_feature_wont_be_returned(self):
        feature_store = self.spark.createDataFrame(
            [
                ["1", dt.date(2020, 1, 1), hashlib.md5("f1".encode()).hexdigest(), "c1f1", None, None],
                ["2", dt.date(2020, 1, 1), hashlib.md5("f1".encode()).hexdigest(), "c2f1", None, None],
                ["1", dt.date(2020, 1, 2), hashlib.md5("f3".encode()).hexdigest(), None, None, "c1f3"],
            ],
            f"{self.__entity.id_column} string, {self.__entity.time_column} date, features_hash string, f1 string, f2 string, f3 string",
        )

        rainbow_table = self.spark.createDataFrame(
            [
                [hashlib.md5("f1".encode()).hexdigest(), ["f1"]],
                [hashlib.md5("f3".encode()).hexdigest(), ["f3"]],
            ],
            ["features_hash", "computed_columns"],
        )

        features_data = self.__filtering_manager.get_latest(feature_store, rainbow_table, ["f1", "f2", "f3"])

        expected_features_data = self.spark.createDataFrame(
            [
                ["1", "c1f1", "c1f3"],
            ],
            f"{self.__entity.id_column} string, f1 string, f3 string",
        )

        self.compare_dataframes(features_data, expected_features_data, [self.__entity.id_column])

    def test_null_returned_correctly(self):
        feature_store = self.spark.createDataFrame(
            [
                ["1", dt.date(2020, 1, 1), hashlib.md5("f1".encode()).hexdigest(), "c1f1", None, None],
                ["2", dt.date(2020, 1, 1), hashlib.md5("f1".encode()).hexdigest(), "c2f1", None, None],
                ["1", dt.date(2020, 1, 2), hashlib.md5("f1`f3".encode()).hexdigest(), None, None, "c1f3"],
            ],
            f"{self.__entity.id_column} string, {self.__entity.time_column} date, features_hash string, f1 string, f2 string, f3 string",
        )

        rainbow_table = self.spark.createDataFrame(
            [
                [hashlib.md5("f1".encode()).hexdigest(), ["f1"]],
                [hashlib.md5("f3".encode()).hexdigest(), ["f3"]],
                [hashlib.md5("f1`f3".encode()).hexdigest(), ["f1", "f3"]],
            ],
            ["features_hash", "computed_columns"],
        )

        features_data = self.__filtering_manager.get_latest(feature_store, rainbow_table, ["f1", "f2", "f3"])

        expected_features_data = self.spark.createDataFrame(
            [
                ["1", None, "c1f3"],
            ],
            f"{self.__entity.id_column} string, f1 string, f3 string",
        )

        self.compare_dataframes(features_data, expected_features_data, [self.__entity.id_column])

    def test_types_are_preserved(self):
        feature_store = self.spark.createDataFrame(
            [
                ["1", dt.date(2020, 1, 1), hashlib.md5("f1".encode()).hexdigest(), 111, None, None],
                ["2", dt.date(2020, 1, 1), hashlib.md5("f1".encode()).hexdigest(), 222, None, None],
                ["1", dt.date(2020, 1, 2), hashlib.md5("f1`f3".encode()).hexdigest(), None, None, 1.12345678910111213],
            ],
            f"{self.__entity.id_column} string, {self.__entity.time_column} date, features_hash string, f1 int, f2 int, f3 double",
        )

        rainbow_table = self.spark.createDataFrame(
            [
                [hashlib.md5("f1".encode()).hexdigest(), ["f1"]],
                [hashlib.md5("f1`f3".encode()).hexdigest(), ["f1", "f3"]],
            ],
            ["features_hash", "computed_columns"],
        )

        features_data = self.__filtering_manager.get_latest(feature_store, rainbow_table, ["f1", "f2", "f3"])

        expected_features_data = self.spark.createDataFrame(
            [
                ["1", None, 1.12345678910111213],
            ],
            f"{self.__entity.id_column} string, f1 int, f3 double",
        )

        self.compare_dataframes(features_data, expected_features_data, [self.__entity.id_column])


if __name__ == "__main__":
    unittest.main()



from pyspark.sql import SparkSession
from pyspark.sql import DataFrame
from pyspark.sql.types import StructType


class EmptyDataFrameCreator:
    def __init__(self, spark: SparkSession):
        self.__spark = spark

    def create(self, schema: StructType) -> DataFrame:
        return self.__spark.createDataFrame(self.__spark.sparkContext.emptyRDD(), schema=schema)



import re
from typing import Dict

from featurestorebundle.feature.FeatureTemplate import FeatureTemplate


class FeaturePattern:
    def __init__(self, feature_template: FeatureTemplate):

        self.__feature_template = feature_template
        self.__placeholders = re.findall(r"{(\w+)}", feature_template.name_template)
        placeholder_translations = {placeholder: f"(?P<{placeholder}>.+)" for placeholder in self.__placeholders}
        self.__pattern = re.compile(feature_template.name_template.format(**placeholder_translations))

    @property
    def feature_template(self):
        return self.__feature_template

    def get_match(self, feature_name: str) -> re.Match:
        return self.__pattern.fullmatch(feature_name)

    def get_groups_as_dict(self, match: re.Match) -> Dict[str, str]:
        return {placeholder: match.group(placeholder) for placeholder in self.__placeholders}



from logging import Logger
from pyspark.sql import SparkSession
from featurestorebundle.delta.TableExistenceChecker import TableExistenceChecker
from featurestorebundle.databricks.feature.writer.DatabricksFeatureStoreMergeConfig import DatabricksFeatureStoreMergeConfig


class DatabricksFeatureStoreDataHandler:
    def __init__(
        self,
        logger: Logger,
        spark: SparkSession,
        table_existence_checker: TableExistenceChecker,
    ):
        self.__logger = logger
        self.__spark = spark
        self.__table_existence_checker = table_existence_checker

    def merge_to_databricks_feature_store(self, full_table_name: str, config: DatabricksFeatureStoreMergeConfig):
        try:
            from databricks import feature_store  # pyre-ignore[21] pylint: disable=import-outside-toplevel

        except ImportError as exception:
            raise Exception("Cannot import Databricks Feature Store, you need to use ML cluster with DBR 8.3+") from exception

        dbx_feature_store_client = feature_store.FeatureStoreClient()
        db_name = full_table_name.split(".")[0]

        self.__spark.sql(f"CREATE DATABASE IF NOT EXISTS {db_name}")

        if not self.__table_exists(dbx_feature_store_client, full_table_name):
            dbx_feature_store_client.create_feature_table(
                name=full_table_name, keys=config.pk_columns, schema=config.data.select(config.pk_columns).schema
            )

        self.__logger.info(f"Writing feature data into Databricks Feature Store table {full_table_name}")

        dbx_feature_store_client.write_table(
            name=full_table_name,
            df=config.data,
            mode="merge",
        )

    def __table_exists(self, fs_client, full_table_name: str):
        if self.__table_exists_in_feature_store(fs_client, full_table_name) and not self.__table_exists_in_hive(full_table_name):
            raise Exception(f"Table {full_table_name} exists in Databricks Feature Store but not in hive")

        if not self.__table_exists_in_feature_store(fs_client, full_table_name) and self.__table_exists_in_hive(full_table_name):
            raise Exception(f"Table {full_table_name} exists in hive but not in Databricks Feature Store")

        if self.__table_exists_in_feature_store(fs_client, full_table_name) and self.__table_exists_in_hive(full_table_name):
            self.__logger.info(f"Databricks Feature Store table {full_table_name} already exists, creation skipped")
            return True

        if not self.__table_exists_in_feature_store(fs_client, full_table_name) and not self.__table_exists_in_hive(full_table_name):
            self.__logger.info(f"Creating new Databricks Feature Store table {full_table_name}")
            return False

        raise Exception(f"Databricks Feature store table {full_table_name} is in inconsistent state")

    def __table_exists_in_feature_store(self, dbx_feature_store_client, full_table_name: str):  # noqa
        try:
            dbx_feature_store_client.get_feature_table(full_table_name)
            return True

        except Exception:  # noqa pylint: disable=broad-except
            return False

    def __table_exists_in_hive(self, full_table_name: str):
        self.__table_existence_checker.exists(full_table_name)



from box import Box
from pyspark.sql import types as t
from daipecore.widgets.Widgets import Widgets
from featurestorebundle.entity.Entity import Entity


class EntityGetter:

    _allowed_id_column_types = ["string", "integer", "long", "short"]
    _types_map = {
        "string": t.StringType(),
        "integer": t.IntegerType(),
        "long": t.LongType(),
        "short": t.ShortType(),
    }

    def __init__(self, entities: Box, widgets: Widgets):
        self.__entities = entities
        self.__widgets = widgets

    def get(self) -> Entity:
        if len(self.__entities) > 1:
            entity_name = self.__widgets.get_value("entity")
        else:
            entity_name = list(self.__entities)[0]

        entity = self.__find_entity_by_name(entity_name)

        return self.__create_entity(entity_name, entity.id_column, entity.id_column_type)

    def get_by_name(self, entity_name: str) -> Entity:
        entity = self.__find_entity_by_name(entity_name)

        return self.__create_entity(entity_name, entity.id_column, entity.id_column_type)

    def __create_entity(self, entity_name: str, id_column: str, id_column_type: str) -> Entity:
        if id_column_type not in self._allowed_id_column_types:
            raise Exception(f"Invalid id column type, allowed types are {self._allowed_id_column_types}")

        return Entity(
            entity_name,
            id_column,
            self._types_map[id_column_type],
            "run_date",
            t.DateType(),
        )

    def __find_entity_by_name(self, entity_name: str) -> Box:
        for entity in self.__entities:
            if entity_name == entity:
                return self.__entities[entity]

        raise Exception(f"Cannot find entity {entity_name} in config")



from typing import Optional
from pyspark.sql import SparkSession
from pyspark.sql import DataFrame
from pyspark.sql import functions as f
from logging import Logger
from featurestorebundle.db.TableNames import TableNames
from featurestorebundle.metadata.reader.MetadataReaderInterface import MetadataReaderInterface


class DeltaPathMetadataReader(MetadataReaderInterface):
    def __init__(
        self,
        logger: Logger,
        table_names: TableNames,
        spark: SparkSession,
    ):
        self.__logger = logger
        self.__spark = spark
        self.__table_names = table_names

    def read(self, entity_name: Optional[str] = None) -> DataFrame:
        entity_name = entity_name or ""
        path = self.__table_names.get_metadata_path(entity_name)

        self.__logger.info(f"Reading metadata from path {path}")

        df = self.__spark.read.format("delta").load(path)

        if entity_name:
            df = df.filter(f.col("entity") == entity_name)

        return df
